# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Base custom error page
"""
from os.path import dirname, join

from ..exceptions import ActionRedirect, RequestBlocked
from ..rules import RuleCallback


class BaseSqreenErrorPage(RuleCallback):
    def __init__(self, *args, **kwargs):
        super(BaseSqreenErrorPage, self).__init__(*args, **kwargs)
        with open(join(dirname(__file__), "sqreen_error_page.html")) as f:
            self.content = f.read()

        rule_data = self.data["values"][0]
        self.rule_type = rule_data["type"]

        if self.rule_type == "custom_error_page":
            self.status_code = int(rule_data["status_code"])
            self.storage.attack_http_code = self.status_code
        elif self.rule_type == "redirection":
            self.redirection_url = rule_data["redirection_url"]
        else:
            raise ValueError("Invalid rule_type %s" % self.rule_type)

    def pre(self, original, *args, **kwargs):
        exception = self._get_exception(*args, **kwargs)

        if not exception:
            return

        if isinstance(exception, RequestBlocked):
            exception.done()
        else:
            raise ValueError(
                "invalid exception type: expected {!r}, got {!r}".format(
                    RequestBlocked, exception.__class__.__name__
                )
            )

        if isinstance(exception, ActionRedirect):
            headers = {"Location": exception.target_url}
            response = self._get_response(exception, "", 303, headers)
            self.record_observation("http_code", "303", 1)
        # Else, exception is either ActionBlock or AttackBlocked.
        elif self.rule_type == "custom_error_page":
            response = self._get_response(
                exception, self.content, self.status_code
            )
            self.record_observation("http_code", str(self.status_code), 1)
        elif self.rule_type == "redirection":
            headers = {"Location": self.redirection_url}
            response = self._get_response(exception, "", 303, headers)
            self.record_observation("http_code", "303", 1)
        else:
            raise ValueError("invalid rule type: {!r}".format(self.rule_type))

        return {"status": "override", "new_return_value": response}



# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
"""Strategy classes for aiohttp."""

from logging import getLogger

from ..middlewares import AioHTTPMiddleware
from .framework import FrameworkStrategy
from .import_hook import ImportHookStrategy

LOGGER = getLogger(__name__)


def get_aiohttp_middleware():
    """Return the aiohttp middleware to be installed.

    In aiohttp 2.2, middlewares are factory functions that return request
    handlers. In aiohttp 2.3, middleware factories (aka old-style middlewares)
    are deprecated in favor of simpler middleware functions, decorated with
    aiohttp.web.middleware. This function returns a suited middleware,
    depending on aiohttp version.
    """
    try:
        from aiohttp.web import middleware  # noqa
    except ImportError:
        LOGGER.debug("New style middlewares are not supported")
        return AioHTTPMiddleware.factory
    else:
        LOGGER.debug("New style middlewares are supported")
        return AioHTTPMiddleware.handle


class AioHTTPInstallStrategy(ImportHookStrategy):
    """Wrap aiohttp.web::Application.freeze to install a custom middleware."""

    def import_hook_callback(self, original):
        def app_freeze(app):
            middleware = get_aiohttp_middleware()
            if middleware in app._middlewares:
                # Prevents inserting the middleware twice.
                LOGGER.warning("Custom middleware already set, skipped")
            else:
                LOGGER.info("Injecting custom middleware")
                app._middlewares.insert(0, middleware)
            return original(app)

        return app_freeze


class AioHTTPHookStrategy(FrameworkStrategy):
    """Hook aiohttp custom middleware."""

    MODULE_NAME = "sqreen.instrumentation.middlewares.aiohttp_middleware"
    HOOK_CLASS = "AioHTTPMiddleware"
    HOOK_METHOD = "handle"

    def __init__(
        self,
        strategy_key,
        observation_queue,
        queue,
        import_hook,
        before_hook_point=None,
    ):
        super(AioHTTPHookStrategy, self).__init__(
            strategy_key,
            observation_queue,
            queue,
            import_hook,
            before_hook_point,
        )
        self.middleware = AioHTTPMiddleware(self, observation_queue, queue)
        self.wrapper = self._apply_middleware

    @staticmethod
    def _apply_middleware(original, middleware):
        return middleware(original)



# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Generic WSGI HTTP Request / Response stuff
"""
import logging
from copy import copy
from itertools import chain

from ..utils import to_unicode_safe
from .base import BaseRequest, BaseResponse
from .ip_utils import get_real_user_ip

try:
    from Cookie import SimpleCookie
except ImportError:
    from http.cookies import SimpleCookie

try:
    from urllib.parse import parse_qs, quote
except ImportError:
    from urlparse import parse_qs
    from urllib import quote


LOGGER = logging.getLogger(__name__)


class WSGIRequest(BaseRequest):
    """ Helper around raw wsgi environ
    """

    def __init__(self, environ):
        super(WSGIRequest, self).__init__()
        self.environ = environ

        form_environ = copy(self.environ)
        form_environ["QUERY_STRING"] = ""

        # TODO: Reactivate reading the body

        self.cookies = self._parse_cookies(form_environ.get("HTTP_COOKIES"))

    @staticmethod
    def _parse_cookies(cookies):
        if not cookies:
            return {}

        cookie = SimpleCookie()
        cookie.load(cookies)

        return {key: cookie[key].coded_value for key in cookie.keys()}

    @property
    def query_params(self):
        """ Return parsed query string from request
        """
        raw_query = self.environ.get("QUERY_STRING", "")
        try:
            return parse_qs(raw_query)
        except Exception:
            LOGGER.warning(
                "Exception while parsing %s", raw_query, exc_info=True
            )
            return {}

    @property
    def form_params(self):
        # TODO: Reactivate reading the body
        return {}

    @property
    def cookies_params(self):
        return self.cookies

    @property
    def query_params_values(self):
        """ Return only query values as a list
        """
        return list(chain.from_iterable(self.query_params.values()))

    @property
    def remote_addr(self):
        """Remote IP address."""
        return self.get_raw_header("REMOTE_ADDR")

    @property
    def raw_client_ip(self):
        return get_real_user_ip(
            self.remote_addr, self.environ.get("HTTP_X_FORWARDED_FOR", "")
        )

    @property
    def hostname(self):
        return to_unicode_safe(self.environ.get("HTTP_HOST", self.environ.get("SERVER_NAME")))

    @property
    def method(self):
        return self.get_raw_header("REQUEST_METHOD")

    @property
    def client_user_agent(self):
        return self.get_raw_header("HTTP_USER_AGENT")

    @property
    def referer(self):
        return self.get_raw_header("HTTP_REFERER")

    @property
    def scheme(self):
        return self.get_raw_header("wsgi.url_scheme")

    @property
    def server_port(self):
        return self.get_raw_header("SERVER_PORT")

    @property
    def remote_port(self):
        return self.get_raw_header("REMOTE_PORT")

    @property
    def path(self):
        return quote(self.environ.get("SCRIPT_NAME", "")) + quote(
            self.environ.get("PATH_INFO", "")
        )

    @property
    def raw_headers(self):
        return self.environ

    @property
    def view_params(self):
        return {}

    @property
    def json_params(self):
        return {}


class WSGIResponse(BaseResponse):
    """Helper around raw WSGI response."""

    def __init__(self, status, response_headers=None):
        self.status = status
        self.headers = response_headers or {}

    @property
    def status_code(self):
        try:
            if isinstance(self.status, str):
                return int(self.status.split(" ", 1)[0])
            elif isinstance(self.status, bytes):
                return int(self.status.split(b" ", 1)[0].decode("latin_1"))
        except Exception:
            LOGGER.debug("cannot parse HTTP status", exc_info=True)
            return None

    @property
    def content_type(self):
        return self.headers.get("Content-Type")

    @property
    def content_length(self):
        try:
            return int(self.headers.get("Content-Length"))
        except (ValueError, TypeError):
            return None



# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Instrumentation helpers
"""
import threading
from functools import wraps

LOCAL = threading.local()


def guard_call(key, callback, *args, **kwargs):
    """ Conditionnaly call callback checking that it cannot be called
    recursively
    """
    # Set guard_call_set if not present, required for multi-threaded env
    if not hasattr(LOCAL, "guard_call_set"):
        LOCAL.guard_call_set = set()

    if key in LOCAL.guard_call_set:
        return {"status": "not_executed"}

    LOCAL.guard_call_set.add(key)

    try:
        output = callback(*args, **kwargs)
        return output
    finally:
        try:
            LOCAL.guard_call_set.remove(key)
        except KeyError:
            pass


def partial(func, *base_args):
    """ A replacement for functools.partial that works
    with instances.
    """

    @wraps(func)
    def newfunc(*args, **kwargs):

        final_args = base_args + args

        return func(*final_args, **kwargs)

    return newfunc


class Proxy(object):
    """ An object proxy helper, forward everything to the proxied object
    except attributes starting with _sqreen
    Comes from http://code.activestate.com/recipes/496741-object-proxying/
    Licensed under PSF (https://en.wikipedia.org/wiki/Python_Software_Foundation_License)
    """

    def __init__(self, obj):
        object.__setattr__(self, "_obj", obj)

    #
    # proxying (special cases)
    #
    def __getattr__(self, name):
        if name.startswith("_sqreen"):
            return getattr(self, name)
        else:
            return getattr(object.__getattribute__(self, "_obj"), name)

    def __delattr__(self, name):
        delattr(object.__getattribute__(self, "_obj"), name)

    def __setattr__(self, name, value):
        if name.startswith("_sqreen"):
            object.__setattr__(self, name, value)
        else:
            setattr(object.__getattribute__(self, "_obj"), name, value)

    def __nonzero__(self):
        return bool(object.__getattribute__(self, "_obj"))

    def __str__(self):
        return str(object.__getattribute__(self, "_obj"))

    def __repr__(self):
        return repr(object.__getattribute__(self, "_obj"))

    #
    # factories
    #
    _special_names = [
        "__abs__",
        "__add__",
        "__and__",
        "__call__",
        "__cmp__",
        "__coerce__",
        "__contains__",
        "__delitem__",
        "__delslice__",
        "__div__",
        "__divmod__",
        "__eq__",
        "__float__",
        "__floordiv__",
        "__ge__",
        "__getitem__",
        "__getslice__",
        "__gt__",
        "__hash__",
        "__hex__",
        "__iadd__",
        "__iand__",
        "__idiv__",
        "__idivmod__",
        "__ifloordiv__",
        "__ilshift__",
        "__imod__",
        "__imul__",
        "__int__",
        "__invert__",
        "__ior__",
        "__ipow__",
        "__irshift__",
        "__isub__",
        "__iter__",
        "__itruediv__",
        "__ixor__",
        "__le__",
        "__len__",
        "__long__",
        "__lshift__",
        "__lt__",
        "__mod__",
        "__mul__",
        "__ne__",
        "__neg__",
        "__oct__",
        "__or__",
        "__pos__",
        "__pow__",
        "__radd__",
        "__rand__",
        "__rdiv__",
        "__rdivmod__",
        "__reduce__",
        "__reduce_ex__",
        "__repr__",
        "__reversed__",
        "__rfloorfiv__",
        "__rlshift__",
        "__rmod__",
        "__rmul__",
        "__ror__",
        "__rpow__",
        "__rrshift__",
        "__rshift__",
        "__rsub__",
        "__rtruediv__",
        "__rxor__",
        "__setitem__",
        "__setslice__",
        "__sub__",
        "__truediv__",
        "__xor__",
        "next",
        "__enter__",
        "__exit__",
    ]

    @classmethod
    def _create_class_proxy(cls, theclass):
        """creates a proxy for the given class"""

        def make_method(name):
            def method(self, *args, **kw):
                return getattr(object.__getattribute__(self, "_obj"), name)(
                    *args, **kw
                )

            return method

        namespace = {}
        for name in cls._special_names:
            if hasattr(theclass, name):
                namespace[name] = make_method(name)
        return type(
            "%s(%s)" % (cls.__name__, theclass.__name__), (cls,), namespace
        )

    def __new__(cls, obj, *args, **kwargs):
        """
        creates an proxy instance referencing `obj`. (obj, *args, **kwargs) are
        passed to this class' __init__, so deriving classes can define an
        __init__ method of their own.
        note: _class_proxy_cache is unique per deriving class (each deriving
        class must hold its own cache)
        """
        try:
            cache = cls.__dict__["_class_proxy_cache"]
        except KeyError:
            cls._class_proxy_cache = cache = {}
        try:
            theclass = cache[obj.__class__]
        except KeyError:
            cache[obj.__class__] = theclass = cls._create_class_proxy(
                obj.__class__
            )
        ins = object.__new__(theclass)
        theclass.__init__(ins, obj, *args, **kwargs)
        return ins



# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Look for known crawlers user-agents
"""
from logging import getLogger

from ..frameworks.pyramid_framework import PyramidRequest, PyramidResponse
from .record_request_context import RecordRequestContext

LOGGER = getLogger(__name__)


class RecordRequestContextPyramid(RecordRequestContext):
    def pre(self, original, request):
        self._store_request(PyramidRequest(request))

    def post(self, *args, **kwargs):
        self.storage.store_response(PyramidResponse(args[1]))
        self._clear_request()



# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Django specific WSGI HTTP Request / Response stuff
"""
from itertools import chain
from logging import getLogger

import pkg_resources

from .base import BaseRequest, BaseResponse

LOGGER = getLogger(__name__)


def drf_preload_data():
    """Detect if request data should be preloaded for Django Rest Framework.

    See https://github.com/encode/django-rest-framework/issues/3951.
    """
    for package in pkg_resources.working_set:
        if package.project_name == "djangorestframework":
            drf_version_info = tuple(map(int, package.version.split(".")[:2]))
            return (3, 3) <= drf_version_info <= (3, 4)
    return False


class DjangoRequest(BaseRequest):

    DRF_PRELOAD_DATA = drf_preload_data()

    def __init__(self, request, view_func, view_args, view_kwargs):
        super(DjangoRequest, self).__init__()
        self.request = request
        self.view_func = view_func
        self.view_args = view_args
        self.view_kwargs = view_kwargs

        # Cache for params
        self._query_params = None
        self._query_params_values = None
        self._form_params = None

        if self.__class__.DEBUG_MODE is None:
            self.__class__.DEBUG_MODE = self.is_debug()

    @property
    def query_params(self):
        if self._query_params is None:
            # Convert django QueryDict to a normal dict with values as list
            self._query_params = dict(self.request.GET.lists())
        return self._query_params

    @property
    def query_params_values(self):
        """ Return only query values as a list
        """
        if self._query_params_values is None:
            self._query_params_values = list(
                chain.from_iterable(self.query_params.values())
            )
        return self._query_params_values

    @property
    def form_params(self):
        if (
            self.DRF_PRELOAD_DATA
            and self.request.method == "POST"
            and not self.request._read_started
        ):
            self.request.body

        if self._form_params is None:
            self._form_params = dict(self.request.POST)
        return self._form_params

    @property
    def cookies_params(self):
        return self.request.COOKIES

    @property
    def remote_addr(self):
        return self.get_raw_header("REMOTE_ADDR")

    @property
    def hostname(self):
        return self.request.get_host()

    @property
    def method(self):
        return self.request.method

    @property
    def referer(self):
        return self.get_raw_header("HTTP_REFERER")

    @property
    def client_user_agent(self):
        return self.get_raw_header("HTTP_USER_AGENT")

    @property
    def route(self):
        """Request route."""
        # The resolver_match attribute is set only set once all request
        # middlewares have been called.
        resolver_match = getattr(self.request, "resolver_match", None)
        return getattr(resolver_match, "view_name", None)

    @property
    def path(self):
        return self.request.path

    @property
    def scheme(self):
        return getattr(self.request, "scheme", None)

    @property
    def server_port(self):
        return self.get_raw_header("SERVER_PORT")

    @property
    def remote_port(self):
        return self.get_raw_header("REMOTE_PORT")

    @property
    def raw_headers(self):
        return self.request.META

    @property
    def view_params(self):
        return self.view_kwargs

    @property
    def json_params(self):
        return {}

    @classmethod
    def is_debug(self):
        try:
            from django.conf import settings

            return getattr(settings, "DEBUG", False)
        except Exception:
            LOGGER.warning("Exception when checking debug mode", exc_info=1)
            return False


class DjangoResponse(BaseResponse):

    def __init__(self, response):
        self.response = response

    @property
    def status_code(self):
        return self.response.status_code

    @property
    def content_type(self):
        return self.response.get("Content-Type")

    @property
    def content_length(self):
        try:
            return int(self.response.get("Content-Length"))
        except (ValueError, TypeError):
            return None



# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Sqreen attack event helpers and placeholder
"""
import traceback
from logging import getLogger

from .config import CONFIG
from .remote_exception import traceback_formatter
from .sanitizer import strip_sensitive_data

LOGGER = getLogger(__name__)


def get_context_payload():
    """ Return attack payload dependent on the context, right now stacktrace.
    """
    return {
        "context": {
            "backtrace": list(traceback_formatter(traceback.extract_stack()))
        }
    }


class Attack(object):
    def __init__(self, payload, rule_name):
        self.payload = payload
        self.rule_name = rule_name

    def to_dict(self):
        result = {}
        rule_payload = self.payload.get("rule", {})
        request_payload = self.payload.get("request", {})
        local_payload = self.payload.get("local", {})
        if "name" in rule_payload:
            result["rule_name"] = rule_payload["name"]
        if "rulespack_id" in rule_payload:
            result["rulespack_id"] = rule_payload["rulespack_id"]
        if "test" in rule_payload:
            result["test"] = rule_payload["test"]
        if "infos" in self.payload:
            result["infos"] = self.payload["infos"]
        if "time" in local_payload:
            result["time"] = local_payload["time"]
        if "remote_ip" in request_payload:
            result["client_ip"] = request_payload["remote_ip"]
        if "request" in self.payload:
            result["request"] = self.payload["request"]
        if "params" in self.payload:
            result["params"] = self.payload["params"]
        if "context" in self.payload:
            result["context"] = self.payload["context"]
        if "headers" in self.payload:
            result["headers"] = self.payload["headers"]
        return result


class RequestRecord(object):
    """Request record objects."""

    VERSION = "20171208"

    def __init__(self, payload):
        self.payload = payload

    def to_dict(self):
        """Export the record as a dict object."""
        result = {"version": self.VERSION}
        if "observed" in self.payload:
            observed_dict = self.payload["observed"]
            result["observed"] = observed_dict
            rulespack = None
            for attack_dict in observed_dict.get("attacks", []):
                rulespack = attack_dict.pop("rulespack_id", None) or rulespack
            for exc_dict in observed_dict.get("sqreen_exceptions", []):
                payload_dict = exc_dict.pop("exception", None)
                if payload_dict:
                    exc_dict["message"] = payload_dict["message"]
                    exc_dict["klass"] = payload_dict["klass"]
                rulespack = exc_dict.pop("rulespack_id", None) or rulespack
            if rulespack:
                result["rulespack_id"] = rulespack
            if "observations" in observed_dict:
                result["observed"]["observations"] = [
                    {"category": cat, "key": key, "value": value, "time": time}
                    for (cat, time, key, value) in observed_dict[
                        "observations"
                    ]
                ]
            if "sdk" in observed_dict:
                result["observed"]["sdk"] = [
                    {"name": entry[0], "time": entry[1], "args": entry[2:]}
                    for entry in observed_dict["sdk"]
                ]
        if "local" in self.payload:
            result["local"] = self.payload["local"]
        if "request" in self.payload:
            request_dict = self.payload["request"]
            result["request"] = request_dict
            if "client_ip" in request_dict:
                result["client_ip"] = request_dict.pop("client_ip")
        else:
            result["request"] = {}
        if "params" in self.payload:
            result["request"]["parameters"] = self.payload["params"]
        if "headers" in self.payload:
            result["request"]["headers"] = self.payload["headers"]

        if CONFIG["STRIP_SENSITIVE_DATA"]:
            result["request"] = strip_sensitive_data(result["request"])

        if "response" in self.payload:
            result["response"] = self.payload["response"]

        return result



# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Record request context
"""

from logging import getLogger

from ..frameworks.wsgi import WSGIRequest
from ..rules import RuleCallback

LOGGER = getLogger(__name__)


class RecordRequestContext(RuleCallback):
    def pre(self, *args, **kwargs):
        self._store_request(WSGIRequest(args[-2]))

    def _store_request(self, request):
        self.storage.store_request(
            request, trace_request=self.performance_monitoring_enabled)
        if self.runner is None or not self.runner.whitelisted_metric:
            return
        whitelist_match = self.storage.get_whitelist_match(self.runner.settings)
        if whitelist_match is not None:
            self.record_observation("whitelisted", whitelist_match, 1)

    def _clear_request(self):
        if self.runner is None:
            self.storage.clear_request(None, None)
        else:
            self.storage.clear_request(
                self.runner.queue, self.runner.observation_queue
            )

    def post(self, *args, **kwargs):
        self._clear_request()

    def failing(self, *args, **kwargs):
        self._clear_request()

    @property
    def whitelisted(self):
        return False



# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Custom error page for Flask
"""
from ..exceptions import RequestBlocked
from .headers_insert import convert_to_str
from .sqreen_error_page import BaseSqreenErrorPage


class SqreenErrorPageFlask(BaseSqreenErrorPage):
    @staticmethod
    def _get_exception(app, exception, *args, **kwargs):
        if not exception:
            return

        # Ignore exception which are not attack blocked
        if not isinstance(exception, RequestBlocked):
            return

        return exception

    @staticmethod
    def _get_response(exception, content, status_code, headers={}):
        from flask import make_response

        response = make_response(content)
        response.status_code = status_code

        for header_name, header_value in convert_to_str(headers.items()):
            response.headers[header_name] = header_value

        return response



# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
"""Binding accessor class."""

import re
import types
from os import getcwd


def flat_keys(iterable, max_iterations=1000):
    """Return the list of keys in iterable and nested iterables."""
    iteration = 0

    keys = []
    remaining_iterables = [iterable]

    seen_iterables = set()

    while len(remaining_iterables) != 0:

        iteration += 1
        # If we have a very big or nested iterable, return False.
        if iteration >= max_iterations:
            break

        iterable = remaining_iterables.pop(0)

        # Protection against recursive objects.
        if id(iterable) in seen_iterables:
            continue

        seen_iterables.add(id(iterable))

        # If we get an iterable, add it to the list of remaining iterables.
        if isinstance(iterable, dict):

            dict_values = iterable.values()
            if len(dict_values) > 0:
                remaining_iterables.extend(list(dict_values))

            dict_keys = iterable.keys()
            if len(dict_keys) > 0:
                keys.extend(dict_keys)

        elif isinstance(iterable, list) and len(iterable) > 0:
            remaining_iterables.extend(iterable)

    return keys


def flat_values(iterable, max_iterations=1000):
    """Return the list of values of iterable and nested iterables."""
    iteration = 0

    values = []
    remaining_iterables = [iterable]

    seen_iterables = set()

    while len(remaining_iterables) != 0:

        iteration += 1
        # If we have a very big or nested iterable, returns False.
        if iteration >= max_iterations:
            break

        iterable = remaining_iterables.pop(0)

        # Protection against recursive objects
        if id(iterable) in seen_iterables:
            continue

        seen_iterables.add(id(iterable))

        # If we get an iterable, add it to the list of remaining iterables.
        if isinstance(iterable, dict):

            dict_values = iterable.values()
            if len(dict_values) > 0:
                remaining_iterables.extend(list(dict_values))

        elif isinstance(iterable, list):
            if len(iterable) > 0:
                remaining_iterables.extend(iterable)

        else:
            values.append(iterable)

    return values


class StringScanner(object):
    """Lexical scanning operations on a string.

    This is a poor man equivalent of Ruby's StringScanner class.
    """

    __slots__ = ("string", "_size", "_pos", "match")

    def __init__(self, string):
        self.string = string
        self._size = len(string)
        self._pos = 0
        self.match = None

    def tell(self):
        """Return the current position in the string."""
        return self._pos

    def scan_string(self, sub):
        """Find the given substring from the current position.

        If the substring is present, the cursor position is updated and the
        substring is returned. Otherwise, this method returns None.

        This is semantically equivalent to calling scan_regex with an exact
        regexp, but faster.
        """
        size = len(sub)
        if self.string[self._pos : self._pos + size] == sub:
            self._pos += size
            return sub

    def scan_regex(self, regex):
        """Match the regex from the current position.

        If there is a match, the cursor position is updated and the match
        object is returned. Otherwise, this method returns None.
        """
        self.match = regex.match(self.string, self._pos)
        if self.match is not None:
            self._pos += len(self.match.group(0))
            return self.match

    def finished(self):
        """True if the cursor is at the end of the string, False otherwise."""
        return self._pos == self._size


class BindingAccessor(object):
    """Binding accessors.

    The class constructor is given a binding accessor expression
    (e.g. "#.args[0]") and _parses it.
    """

    __slots__ = (
        "path",
        "expression",
        "transformation",
        "_transformation_func",
        "_scanner",
    )

    def __init__(self, expression):
        self.path = []
        self.expression = expression
        self.transformation = None
        self._transformation_func = None
        self._scanner = None
        self._parse(expression)

    def _parse(self, expression):
        """Parse a binding accessor expression.

        It is internally converted to a series of componentss that are later
        resolved.
        """
        # If there is a transformation, set it and remove it
        # (e.g. "expression | transformation").
        expression = self._parse_transformation(expression)

        self._scanner = StringScanner(expression)
        while not self._scanner.finished():
            start_pos = self._scanner.tell()

            # Check for scalar values first.
            scalar = self._scan_scalar()
            if scalar:
                self.path.append(scalar)
                return

            # If we are at the beginning of an expression, a variable is
            # allowed.
            if start_pos == 0:
                self._scan_push_variable()
            else:
                self._scan_push_attribute()
            self._scan_push_indexes()

            # Remove potential dot.
            self._scanner.scan_string(".")

            if start_pos == self._scanner.tell():
                raise ValueError("parsing error, _parser is stuck")

        # Delete the scanner instance.
        self._scanner = None

    _TRANSFORMATION_REGEX = re.compile(r"\|[ \w]*$")
    _TRANSFORMATION_FUNCS = {
        "len": len,
        "flat_keys": flat_keys,
        "flat_values": flat_values,
    }

    def _parse_transformation(self, expression):
        """Try to parse a transformation out of an expression."""
        match = self._TRANSFORMATION_REGEX.search(expression)
        if match is None:
            return expression
        len_suffix = len(match.group(0))
        self.transformation = expression[-len_suffix + 1 :].strip()

        # Check that the transformation name is defined.
        if self.transformation not in self._TRANSFORMATION_FUNCS:
            raise ValueError(
                "unknown transformation {!r} for expression {!r}".format(
                    self.transformation, expression
                )
            )

        self._transformation_func = self._TRANSFORMATION_FUNCS[
            self.transformation
        ]
        return expression[:-len_suffix].rstrip()

    _SQREEN_VARIABLE_REGEX = re.compile(r"#\.(\w+)")
    _PYTHON_IDENTIFIER_REGEX = re.compile(r"[a-zA-Z_](?:\w)*")

    def _scan_push_variable(self):
        """Try to parse a variable and push it to the components list.

        Do nothing if the scanner cannot parse a variable or a Python
        identifier.
        """
        if self._scanner.scan_regex(self._SQREEN_VARIABLE_REGEX):
            self.path.append(
                {
                    "name": self._scanner.match.group(1),
                    "kind": "sqreen-variable",
                }
            )
        elif self._scanner.scan_regex(self._PYTHON_IDENTIFIER_REGEX):
            self.path.append(
                {"name": self._scanner.match.group(), "kind": "variable"}
            )

    def _scan_push_attribute(self):
        """Try to parse an attribute and push it to the components list.

        Do nothing of the scanner cannot parse a Python identifier.
        """
        if self._scanner.scan_regex(self._PYTHON_IDENTIFIER_REGEX):
            self.path.append(
                {"name": self._scanner.match.group(), "kind": "attribute"}
            )

    _INTEGER_REGEX = re.compile(r"\d+")
    _STRING_REGEX = re.compile(r"'((?:\\.|[^\\'])*)'")

    def _scan_scalar(self):
        """Scan a scalar value and return the corresponding components.

        Return None if the scanner cannot parse an integer or a string.
        """
        if self._scanner.scan_regex(self._INTEGER_REGEX):
            return {
                "value": int(self._scanner.match.group()),
                "kind": "integer",
            }
        elif self._scanner.scan_regex(self._STRING_REGEX):
            return {"value": self._scanner.match.group(1), "kind": "string"}

    def _scan_push_indexes(self):
        """Scan a sequence of indexes and push them to the components list."""
        while self._scanner.scan_string("[") is not None:
            scalar = self._scan_scalar()
            if scalar is None:
                raise ValueError("invalid index")
            if self._scanner.scan_string("]") is None:
                raise ValueError("unfinished index")
            self.path.append({"index": scalar["value"], "kind": "index"})

    def resolve(
        self,
        binding=None,
        global_binding=None,
        framework=None,
        instance=None,
        arguments=None,
        kwarguments=None,
        cbdata=None,
        return_value=None,
    ):
        """Given a context, resolve the expression and return the value."""
        if binding is None:
            binding = {}
        if global_binding is None:
            global_binding = {}

        env = [
            framework,
            instance,
            arguments,
            kwarguments,
            cbdata,
            return_value,
        ]
        value = None
        for component in self.path:
            value = self._resolve_component(
                value, component, binding, global_binding, env
            )
        self._validate_value(value)

        if self.transformation is not None:
            value = self._transformation_func(value)

        return value

    def _resolve_component(
        self, value, component, binding, global_binding, env
    ):
        """Resolve a component.

        Simple expressions (static strings, integers, indexes and
        attributes) are directly resolved. Others are dispatched to specialized
        methods.
        """
        component_kind = component["kind"]
        if component_kind == "string" or component_kind == "integer":
            return component["value"]
        elif component_kind == "variable":
            return self._resolve_variable(
                component["name"], binding, global_binding
            )
        elif component_kind == "index":
            return value[component["index"]]
        elif component_kind == "attribute":
            return getattr(value, component["name"])
        elif component_kind == "sqreen-variable":
            return self._resolve_sqreen_variable(component["name"], *env)
        else:
            raise ValueError(
                "invalid component kind {!r}".format(component_kind)
            )

    def _resolve_sqreen_variable(
        self,
        what,
        framework,
        instance,
        arguments,
        kwarguments,
        cbdata,
        return_value,
    ):
        """Resolve sqreen-variables (the ones starting with #.).

        Fall back on the framework object if the value is not a special
        sqreen-variable. Return None if the framework is None.
        """
        if what == "data":
            return cbdata
        elif what == "rv":
            return return_value
        elif what == "args" or what == "cargs":
            return arguments
        elif what == "kwargs":
            return kwarguments
        elif what == "inst":
            return instance
        elif what == "cwd":
            return getcwd()
        elif framework is not None:
            return getattr(framework, what)

    def _resolve_variable(self, variable_name, binding, global_binding):
        """Resolve a general variable name.

        Search in local context first, then in general context.
        """
        if variable_name in binding:
            return binding[variable_name]
        elif variable_name in global_binding:
            return global_binding[variable_name]
        else:
            raise NameError(
                "name {!r} was not found in bindings".format(variable_name)
            )

    def _validate_value(self, value):
        """Raise ValueError if the value is a method or a function."""
        if isinstance(value, (types.FunctionType, types.MethodType)):
            raise ValueError("invalid return value {!r}".format(value))

    def __repr__(self):
        return "{}({!r})".format(self.__class__.__name__, self.expression)



# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Various utils
"""
import json
import sys
from functools import WRAPPER_ASSIGNMENTS
from inspect import isclass
from logging import getLogger
from operator import methodcaller

from ._vendors.ipaddress import (
    ip_address as _ip_address,
    ip_network as _ip_network,
)

LOGGER = getLogger(__name__)
PYTHON_VERSION = sys.version_info[0]


if PYTHON_VERSION == 2:
    ALL_STRING_CLASS = basestring  # noqa
    STRING_CLASS = str
    UNICODE_CLASS = unicode  # noqa
elif PYTHON_VERSION == 3:
    ALL_STRING_CLASS = str
    STRING_CLASS = str
    UNICODE_CLASS = str
NONE_TYPE = type(None)


HAS_ASYNCIO = sys.version_info >= (3, 4)


def is_string(value):
    """ Check if a value is a valid string, compatible with python 2 and python 3

    >>> is_string('foo')
    True
    >>> is_string(u'✌')
    True
    >>> is_string(42)
    False
    >>> is_string(('abc',))
    False
    """
    return isinstance(value, ALL_STRING_CLASS)


def is_unicode(value):
    """ Check if a value is a valid unicode string, compatible with python 2 and python 3

    >>> is_unicode(u'foo')
    True
    >>> is_unicode(u'✌')
    True
    >>> is_unicode(b'foo')
    False
    >>> is_unicode(42)
    False
    >>> is_unicode(('abc',))
    False
    """
    return isinstance(value, UNICODE_CLASS)


def to_latin_1(value):
    """ Return the input string encoded in latin1 with replace mode for errors
    """
    return value.encode("latin-1", "replace")


def is_json_serializable(value):
    """ Check that a single value is json serializable
    """
    return isinstance(value, (ALL_STRING_CLASS, NONE_TYPE, bool, int, float))


def update_wrapper(wrapper, wrapped):
    """ Update wrapper attribute to make it look like wrapped function.
    Don't use original update_wrapper because it can breaks if wrapped don't
    have all attributes.
    """
    for attr in WRAPPER_ASSIGNMENTS:
        if hasattr(wrapped, attr):
            setattr(wrapper, attr, getattr(wrapped, attr))
    return wrapper


def to_unicode_safe(value):
    """ Safely convert a value to a unicode string.
    """
    if value is None:
        return value
    # If value is a byte string (string without encoding)
    # Try to decode it as unicode, this operation will
    # always succeed because non UTF-8 characters will
    # get replaced by the UTF-8 replacement character.
    if isinstance(value, bytes):
        return value.decode("utf-8", errors="replace")
    # If the value is not already a unicode string
    # Try to convert it to a unicode string
    # by calling the standard Python __str__ method on
    # the value.
    elif not is_unicode(value):
        return UNICODE_CLASS(value)
    # Value is already a unicode string
    return value


###
# Raven configuration
###


def _raven_ignoring_handler(logger, *args, **kwargs):
    """ Ignore all logging messages from sqreen.* loggers, effectively
    disabling raven to log sqreen log messages as breadcrumbs
    """
    try:
        if logger.name.startswith("sqreen"):
            return True
    except Exception:
        LOGGER.warning("Error in raven ignore handler", exc_info=True)


def configure_raven_breadcrumbs():
    """ Configure raven breadcrumbs logging integration if raven is present
    """
    try:
        from raven import breadcrumbs
    except ImportError:
        return

    # Register our logging handler to stop logging sqreen log messages
    # as breadcrumbs
    try:
        breadcrumbs.register_logging_handler(_raven_ignoring_handler)
    except Exception:
        LOGGER.warning("Error while configuring breadcrumbs", exc_info=True)


###
# JSON Encoder
###


def qualified_class_name(obj):
    """ Return the full qualified name of the class name of obj in form of
    `full_qualified_module.class_name`
    """
    if isclass(obj):
        instance_class = obj
    else:
        instance_class = obj.__class__

    return ".".join([instance_class.__module__, instance_class.__name__])


def django_user_conversion(obj):
    """ Convert a Django user either by returning USERNAME_FIELD or convert
    it to str.
    """
    if hasattr(obj, "USERNAME_FIELD"):
        return getattr(obj, getattr(obj, "USERNAME_FIELD"), None)
    else:
        return str(obj)


class CustomJSONEncoder(json.JSONEncoder):
    def __init__(self, *args, **kwargs):
        super(CustomJSONEncoder, self).__init__(*args, **kwargs)

        self.mapping = {}

        self.mapping["bson.objectid.ObjectId"] = str
        self.mapping[
            "django.contrib.auth.models.AbstractUser"
        ] = django_user_conversion

        # Convert datetime to isoformat, compatible with Node Date()
        self.mapping["datetime.datetime"] = methodcaller("isoformat")

        self.mapping["sqreen._vendors.ipaddress.IPv4Address"] = str
        self.mapping["sqreen._vendors.ipaddress.IPv6Address"] = str

    def default(self, obj):
        """ Return the repr of unkown objects
        """
        if type(obj) == type:
            instance_class = obj
        else:
            instance_class = obj.__class__

        # Manually do isinstance without needed to have a reference to the class
        for klass in instance_class.__mro__:
            qualified_name = qualified_class_name(klass)

            if qualified_name in self.mapping:
                try:
                    return self.mapping[qualified_name](obj)
                except Exception:
                    msg = "Error converting an instance of type %r"
                    LOGGER.warning(msg, obj.__class__, exc_info=True)

        # If we don't, or if we except, fallback on repr
        return repr(obj)


def ip_address(address):
    return _ip_address(UNICODE_CLASS(address))


def ip_network(address, strict=True):
    return _ip_network(UNICODE_CLASS(address), strict=strict)



# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Custom error page for Django new style middlewares
"""
from ..exceptions import RequestBlocked
from .sqreen_error_page_django import SqreenErrorPageDjango


class SqreenErrorPageDjangoNewStyle(SqreenErrorPageDjango):
    @staticmethod
    def _get_exception(request, exc, *args, **kwargs):
        if not exc:
            return

        if not isinstance(exc, RequestBlocked):
            return

        return exc



# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Callbacks rules classes and helpers
"""
import logging
from datetime import datetime
from functools import wraps

from .condition_evaluator import ConditionEvaluator, is_condition_empty
from .constants import LIFECYCLE_METHODS
from .remote_exception import raw_traceback_formatter, traceback_formatter
from .runtime_storage import runtime

LOGGER = logging.getLogger(__name__)


class BaseCallback(object):
    """ Base class for callbacks.

    The hook_name is the path to the hook, it could be either a module, like
    "package.module", or a specific class "package.module::Class".
    The hook_name is the name of the function to hook_on, it's relative to
    the hook_module, for example with a hook_module equal to
    "package.module::Class" and a hook_path equal to "method", we will
    hook on the method named "method" of a class named "Class" in the module
    named "package.module"
    """

    SUPPORTS_BUDGET = False  # override constant in subclasses supporting it

    def __init__(self, hook_module, hook_name, strategy=None):
        self.hook_module = hook_module
        self.hook_name = hook_name
        self.strategy = strategy

    @property
    def supports_budget(self):
        return self.SUPPORTS_BUDGET


def check_condition_wrapper(rule_data, function, condition, lifecycle, storage=runtime):
    """ Wrapper that will check lifecycle method pre-condition before
    calling it.
    If pre-conditions are true, call the lifecycle method, otherwise
    return None.
    """

    @wraps(function)
    def wrapped(*args, **kwargs):
        """ Wrapper around lifecycle method
        """

        # Copy args to work on them
        binding_args = list(args)

        # Compute return value depending on the lifecycle method we wrap
        return_value = False
        if lifecycle in ("post", "failing"):
            # Args[1/2] is either the return value of the hooked point
            # or the exception
            return_value = binding_args.pop(1)

        original = binding_args.pop(0)

        binding_eval_args = {
            "binding": locals(),
            "global_binding": globals(),
            "framework": storage.get_current_request(),
            "instance": original,
            "arguments": storage.get_current_args(binding_args),
            "cbdata": rule_data,
            "return_value": return_value,
        }

        # Check the pre condition
        condition_result = condition.evaluate(**binding_eval_args)

        LOGGER.debug(
            "ConditionWrapper (%s) result for %s: %s",
            condition,
            function,
            condition_result,
        )

        if condition_result in (False, None):
            return None

        # Execute the hook otherwise with the original args
        return function(*args, **kwargs)

    wrapped.__wrapped__ = function

    return wrapped


def call_count_wrapper(
    function, lifecycle, call_count_interval, observation_key, callback
):
    """ Wrapper around lifecycle methods that record number of calls
    """

    @wraps(function)
    def wrapped(*args, **kwargs):
        """ Record the number of calls for this callback lifecycle method.
        Buffer the number in the callback itself (self.call_counts) and record
        an observation every X times, X being the field call_count_interval of
        the rule.
        """
        current_count = callback.call_counts[lifecycle]

        if current_count + 1 == call_count_interval:
            callback.record_observation(
                "sqreen_call_counts", observation_key, call_count_interval
            )
            callback.call_counts[lifecycle] = 0
        else:
            callback.call_counts[lifecycle] += 1

        return function(*args, **kwargs)

    wrapped.__wrapped__ = function

    return wrapped


def perf_monitoring_wrapper(function, name, storage):
    """ Wrapper that will monitor the performance of the rule callback.
    It will time the execution of the callback and send it as an
    observation metric.

    The wrapper should be called first before every others.
    """

    @wraps(function)
    def wrapped(*args, **kwargs):
        """ Time the callback execution.
        """
        with storage.trace(name):
            return function(*args, **kwargs)

    wrapped.__wrapped__ = function

    return wrapped


class RuleCallback(BaseCallback):

    SECTIONS = ("request", "response", "params", "headers", "context")

    def __init__(
        self,
        hook_module,
        hook_name,
        rule_name,
        rulespack_id,
        block,
        test,
        runner,
        strategy_cls=None,
        data=None,
        conditions=None,
        callbacks=None,
        call_count_interval=0,
        metrics=None,
        payload_sections=None,
        storage=runtime,
        attack_render_info=None
    ):
        super(RuleCallback, self).__init__(
            hook_module, hook_name, strategy_cls
        )
        self.rule_name = rule_name
        self.rulespack_id = rulespack_id

        if data is None:
            data = {}
        self.data = data
        self.block = block
        self.test = test
        self.runner = runner
        self.metrics = metrics
        self.storage = storage

        if conditions is None:
            conditions = {}
        self.conditions = conditions

        if callbacks is None:
            callbacks = []
        self.callbacks = callbacks
        self.attack_render_info = attack_render_info

        # Callbacks
        self.call_count_interval = call_count_interval
        self.call_counts = {"pre": 0, "post": 0, "failing": 0}

        self._apply_conditions()
        self._apply_call_count()
        self._apply_perf_monitoring()

        # Payload sections.
        if payload_sections is None:
            payload_sections = self.SECTIONS
        self.payload_sections = payload_sections

    @classmethod
    def from_rule_dict(cls, rule_dict, runner, storage=runtime):
        """ Return a RuleCallback based on a rule dict
        """
        return cls(
            hook_module=rule_dict["hookpoint"]["klass"],
            hook_name=rule_dict["hookpoint"]["method"],
            rule_name=rule_dict["name"],
            rulespack_id=rule_dict["rulespack_id"],
            strategy_cls=rule_dict["hookpoint"].get("strategy", "import_hook"),
            attack_render_info=rule_dict["hookpoint"].get("attack_render_info", {}),
            block=rule_dict["block"],
            test=rule_dict["test"],
            data=rule_dict.get("data"),
            conditions=rule_dict.get("conditions"),
            callbacks=rule_dict.get("callbacks"),
            call_count_interval=rule_dict.get("call_count_interval", 0),
            runner=runner,
            metrics=rule_dict.get("metrics", []),
            payload_sections=rule_dict.get("payload"),
            storage=storage,
        )

    @classmethod
    def get_budget(cls, kwargs):
        if isinstance(kwargs, dict):
            return kwargs.pop("__sqreen_budget", None)
        return None

    @property
    def performance_monitoring_enabled(self):
        """ Returns True if the performance monitoring is enabled for this rule.
        """
        if self.runner is not None:
            return self.runner.performance_metrics and self.runner.performance_metrics_settings.enabled
        return False

    def _apply_conditions(self):
        """ Wrap each lifecycle methods if the Rule define them and if we have
        conditions for them.
        """
        for lifecycle in LIFECYCLE_METHODS.values():
            conditions = self.conditions.get(lifecycle)

            if not is_condition_empty(conditions) and hasattr(self, lifecycle):
                conditions = ConditionEvaluator(conditions)

                # Wrap the lifecycle method
                lifecycle_method = getattr(self, lifecycle)
                wrapped = check_condition_wrapper(
                    self.data, lifecycle_method, conditions, lifecycle,
                    storage=self.storage
                )
                setattr(self, lifecycle, wrapped)

    def _apply_call_count(self):
        # Only count calls if call_count_interval is > 0
        if self.call_count_interval == 0:
            return

        for lifecycle in LIFECYCLE_METHODS.values():
            if hasattr(self, lifecycle):
                lifecycle_method = getattr(self, lifecycle)

                observation_key = "%s/%s/%s" % (
                    self.rulespack_id,
                    self.rule_name,
                    lifecycle,
                )

                wrapped = call_count_wrapper(
                    lifecycle_method,
                    lifecycle,
                    self.call_count_interval,
                    observation_key,
                    self,
                )
                setattr(self, lifecycle, wrapped)

    def _apply_perf_monitoring(self):
        # Only enable the performance monitoring if the feature flag is
        if not self.performance_monitoring_enabled:
            return

        for lifecycle in LIFECYCLE_METHODS.values():
            if hasattr(self, lifecycle):
                lifecycle_method = getattr(self, lifecycle)

                metric_name = "sq.{}.{}".format(self.rule_name, lifecycle)

                # Register the performance metric if it does not exist yet
                if self.runner is not None \
                        and not self.runner.metrics_store.is_registered(metric_name):
                    self.runner.metrics_store.register_metric(
                        metric_name, "Binning",
                        self.runner.performance_metrics_settings.period,
                        base=self.runner.performance_metrics_settings.base,
                        unit=self.runner.performance_metrics_settings.unit
                    )

                wrapped = perf_monitoring_wrapper(
                    lifecycle_method,
                    metric_name,
                    self.storage
                )
                setattr(self, lifecycle, wrapped)

    def exception_infos(self, infos={}):
        """ Returns additional infos in case of exception
        """
        return {"rule_name": self.rule_name, "rulespack_id": self.rulespack_id}

    def record_attack(self, infos=None, at=None):
        """Record an attack."""
        if at is None:
            at = datetime.utcnow()
        payload = {
            "infos": infos,
            "rulespack_id": self.rulespack_id,
            "rule_name": self.rule_name,
            "test": self.test,
            "time": at,
        }
        if "context" in self.payload_sections:
            current_request = self.storage.get_current_request()
            if current_request:
                payload["backtrace"] = list(current_request.raw_caller)
        LOGGER.debug("Observed attack %r", payload)
        self.storage.observe("attacks", payload, self.payload_sections)

    def record_observation(self, metric_name, key, value, at=None):
        """Record a metric observation."""

        self.global_record_observation(metric_name, key, value, at, storage=self.storage)

    @classmethod
    def global_record_observation(cls, metric_name, key, value, at=None, storage=runtime):
        """Record a metric observation"""

        if at is None:
            at = datetime.utcnow()
        payload = (metric_name, at, key, value)
        LOGGER.debug("Observed metric %r", payload)
        storage.observe("observations", payload, report=False)

    def record_exception(self, exc, exc_info, stack=None, infos=None, at=None):
        """Record an exception."""
        if infos is None:
            infos = {}
        if at is None:
            at = datetime.utcnow()
        # Try to recover some infos from the exception if it's a
        # SqreenException.
        try:
            infos["exception"] = exc.exception_infos()
        except Exception:
            pass
        bt = raw_traceback_formatter(exc_info[2])
        if stack is not None:
            bt = traceback_formatter(stack) + bt
        payload = {
            "message": str(exc_info[1]),
            "klass": exc_info[0].__name__,
            "infos": infos,
            "rule_name": self.rule_name,
            "test": self.test,
            "time": at,
            "backtrace": bt,
        }
        LOGGER.debug("Observed exception %r", payload)
        self.storage.observe("sqreen_exceptions", payload, self.payload_sections)

    @property
    def lifecycle_methods(self):
        return [
            lifecycle
            for lifecycle in LIFECYCLE_METHODS.values()
            if hasattr(self, lifecycle)
        ]

    @property
    def whitelisted(self):
        """ Return True if the callback should be skipped (whitelisted
        request), False otherwise
        """
        if self.runner is None:
            return False
        return self.storage.get_whitelist_match(self.runner.settings) is not None

    def __repr__(self):
        return "%s(rule_name=%r)" % (self.__class__.__name__, self.rule_name)



# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Flask specific WSGI HTTP Request / Response stuff
"""

from itertools import chain
from logging import getLogger
from traceback import format_stack

from ..utils import to_unicode_safe
from .base import BaseRequest, BaseResponse

try:
    # Python 2
    import urlparse
except ImportError:
    import urllib.parse as urlparse


LOGGER = getLogger(__name__)


class FlaskRequest(BaseRequest):
    def __init__(self, request):
        super(FlaskRequest, self).__init__()
        self.request = request

        # Convert flask QueryDict to a normal dict with values as list
        self.converted_get = dict(self.request.args.lists())

        # Cache for params
        self._query_params_values = None
        self._form_params = None
        self._cookies_params = None

        if self.__class__.DEBUG_MODE is None:
            self.__class__.DEBUG_MODE = self.is_debug()

    @property
    def query_params(self):
        return self.converted_get

    @property
    def query_params_values(self):
        """ Return only query values as a list
        """
        if self._query_params_values is None:
            self._query_params_values = list(
                chain.from_iterable(self.converted_get.values())
            )
        return self._query_params_values

    @property
    def form_params(self):
        if self._form_params is None:
            self._form_params = dict(self.request.form)
        return self._form_params

    @property
    def cookies_params(self):
        if self._cookies_params is None:
            self._cookies_params = dict(self.request.cookies)
        return self._cookies_params

    @property
    def json_params(self):
        return self.request.get_json(silent=True)

    @property
    def remote_addr(self):
        """Remote IP address."""
        return self.get_raw_header("REMOTE_ADDR")

    @property
    def hostname(self):
        parsed = urlparse.urlparse(self.request.url_root)
        return parsed.netloc

    @property
    def method(self):
        return to_unicode_safe(self.request.method)

    @property
    def referer(self):
        return self.request.referrer

    @property
    def client_user_agent(self):
        return to_unicode_safe(self.request.user_agent.string)

    @property
    def route(self):
        """Request route."""
        url_rule = getattr(self.request, "url_rule", None)
        # If a custom rule_class is used in the app, the rule attribute might
        # be missing.
        return getattr(url_rule, "rule", None)

    @property
    def path(self):
        return self.request.path

    @property
    def scheme(self):
        return to_unicode_safe(self.request.scheme)

    @property
    def server_port(self):
        return self.get_raw_header("SERVER_PORT")

    @property
    def remote_port(self):
        return self.get_raw_header("REMOTE_PORT")

    @property
    def raw_headers(self):
        return self.request.environ

    @property
    def caller(self):
        return format_stack()

    @property
    def view_params(self):
        return self.request.view_args

    @staticmethod
    def is_debug():
        try:
            from flask import current_app

            return current_app.config.get("DEBUG", False)
        except Exception:
            LOGGER.warning("Exception when checking debug mode", exc_info=1)
            return False


class FlaskResponse(BaseResponse):

    def __init__(self, response):
        self.response = response

    @property
    def status_code(self):
        return self.response.status_code

    @property
    def content_type(self):
        return self.response.headers.get("Content-Type")

    @property
    def content_length(self):
        try:
            return int(self.response.headers.get("Content-Length"))
        except (ValueError, TypeError):
            return None



# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Sqreen Python agent thread composition
"""
import atexit
import json
import os
import random
import sys
import threading
import time
import traceback
from copy import copy
from logging import getLogger
from threading import Event, Thread

from .__about__ import __version__
from ._vendors.urllib3.exceptions import HTTPError
from .config import CONFIG
from .constants import CHANGELOG_URL, COMPATIBILITY_URL, STATUS_URL
from .deliverer import get_deliverer
from .exceptions import UnsupportedFrameworkVersion, UnsupportedPythonVersion
from .frameworks import django_framework, flask_framework, pyramid_framework
from .http_client import Urllib3Connection
from .instrumentation import Instrumentation
from .log import configure_root_logger
from .metrics import MetricsStore
from .remote_command import RemoteCommand
from .remote_exception import RemoteException
from .runner import (
    MAX_OBS_QUEUE_LENGTH,
    MAX_QUEUE_LENGTH,
    CappedQueue,
    Runner,
    RunnerSettings,
    RunnerStop,
    process_initial_commands,
)
from .runtime_infos import (
    RuntimeInfos,
    get_parent_cmdline,
    get_process_cmdline,
)
from .session import InvalidToken, Session
from .signature import InvalidSignature
from .utils import HAS_ASYNCIO, configure_raven_breadcrumbs

if HAS_ASYNCIO:
    from .frameworks import aiohttp_framework


RUNNER_THREAD = None
RUNNER_THREAD_PID = None
RUNNER_THREAD_RESTARTING = False
RUNNER_LOCK = threading.RLock()

# Graceful timeout, max time waiting for the runner thread to exit before exiting
try:
    GRACEFUL_TIMEOUT = int(os.getenv("SQREEN_GRACEFUL_TIMEOUT", 5))
except ValueError:
    GRACEFUL_TIMEOUT = 5
GRACEFUL_FAIL_MSG = "Sqreen thread didn't exit after %s seconds"

LOGGER = getLogger(__name__)


def uwsgi_version_info():
    """Return uWSGI version tuple if run in uWSGI, None otherwise."""
    # Import must be delayed inside the function, because the uwsgi module is
    # not known by sqreen-start.
    try:
        import uwsgi
    except ImportError:
        return None
    else:
        return uwsgi.version_info


def should_start_runner_thread():
    """This function contains the logic used to check if the thread
    should be restarted. Beware, this function *must not* call any hooks.
    """
    global RUNNER_THREAD
    global RUNNER_THREAD_PID

    runner_stopped = False
    if RUNNER_THREAD is not None \
            and RUNNER_THREAD.runner is not None \
            and RUNNER_THREAD.runner.stop is True:
        runner_stopped = True

    return RUNNER_THREAD_PID != os.getpid() and not runner_stopped


def before_hook_point():
    global RUNNER_THREAD
    global RUNNER_THREAD_PID
    global RUNNER_THREAD_RESTARTING

    # /!\ The order of execution in this function must not change
    # without a careful review. /!\
    if should_start_runner_thread():
        # It is cheap to first call should_start_runner_thread instead
        # of just locking everytime.
        with RUNNER_LOCK:
            # If we are in a recursive call, do nothing
            if RUNNER_THREAD_RESTARTING is True:
                LOGGER.debug("The runner thread is already restarting")
                return
            # If a concurrent thread has already started the thread, do nothing
            if not should_start_runner_thread():
                LOGGER.info("The runner thread has already been started")
                return
            # Else, restart the thread
            try:
                RUNNER_THREAD_RESTARTING = True
                runner = RUNNER_THREAD.replace()
                runner.start()

                RUNNER_THREAD = runner
                RUNNER_THREAD_PID = os.getpid()

                LOGGER.info(
                    "Sucessfully started runner thread for pid: %d", os.getpid()
                )
            finally:
                RUNNER_THREAD_RESTARTING = False


def _dump_thread_stacks():
    """ Try to dump all the currently running thread stacks
    """
    try:
        msg = ["Threads stacks \n"]
        thread_name_map = {
            thread.ident: thread.name for thread in threading.enumerate()
        }

        for thread_id, stack in sys._current_frames().items():
            msg.append(
                "# Thread: id '%r', name %r"
                % (thread_id, thread_name_map[thread_id])
            )
            msg.append("".join(traceback.format_stack(stack)))

        LOGGER.debug("\n".join(msg))
    except Exception:
        LOGGER.exception("Something happened")


BLACKLISTED_PARENT_COMMANDS = (
    "ipython",
    "celery worker",
    "rq worker",
    "manage.py shell",
    "pshell",
)
BLACKLISTED_COMMANDS = BLACKLISTED_PARENT_COMMANDS + ("newrelic-admin",)

INSTRUMENTATION_TIMEOUT = 5


def get_session(url, token, app_name=None, proxy_url=None):
    """ Create a connection with the endpoint URL and a session. Returns only
    the session.
    """
    LOGGER.warning(
        "Connection to %s%s",
        url,
        " using proxy %s" % proxy_url if proxy_url else "",
    )
    con = Urllib3Connection(url, proxy_url)
    session = Session(con, token, app_name)
    return session


def post_app_sqreen_exception(data):
    """Post a Sqreen exception happening at application level.
    This function does not need a session.

    This function is used to report instrumentation errors. It should never
    fail.
    """
    token = CONFIG["TOKEN"]
    app_name = CONFIG["APP_NAME"]
    if not token:
        return

    try:
        session = get_session(
            CONFIG["URL"], token, app_name, proxy_url=CONFIG["PROXY_URL"]
        )
        session.post_app_sqreen_exception(data)
    except Exception:
        pass


def start():
    """ Start the background thread and start protection
    """
    # Check if the agent is not disabled first
    try:
        disabled = int(os.getenv("SQREEN_DISABLE", 0))
    except ValueError:
        disabled = None

    # Configure logging
    configure_root_logger(CONFIG["LOG_LEVEL"], CONFIG["LOG_LOCATION"])

    # Retrieve the command used to launch the process
    command = get_process_cmdline()

    # Retrieve the parent command
    parent_command = get_parent_cmdline()

    # Check if we shouldn't launch ourselves
    for blacklisted_command in BLACKLISTED_COMMANDS:
        if blacklisted_command in command:
            LOGGER.debug(
                "Sqreen agent is disabled when running %s.",
                blacklisted_command,
            )
            return
    for blacklisted_parent_command in BLACKLISTED_PARENT_COMMANDS:
        if blacklisted_parent_command in parent_command:
            LOGGER.debug(
                "Sqreen agent is disabled when running %s.",
                blacklisted_parent_command,
            )
            return
    if hasattr(sys, "argv") and len(sys.argv) >= 2 and sys.argv[1] == "test":
        LOGGER.debug("Sqreen agent is disabled when running tests.")
        return

    if disabled:
        LOGGER.debug("Sqreen agent is disabled.")
        return

    try:
        runtime_infos = RuntimeInfos().all()
    except UnsupportedFrameworkVersion as exception:
        msg = (
            "%s version %s is not supported in this agent version.\n"
            "Sqreen agent is disabled, you're not protected.\n"
            "Documentation can be found at %s.\n"
            "Changelog can be found at %s."
        )
        LOGGER.critical(
            msg,
            exception.framework.title(),
            exception.version,
            COMPATIBILITY_URL,
            CHANGELOG_URL,
        )
        # Alter os.environ?
        return
    except UnsupportedPythonVersion as exception:
        msg = (
            "Python version %s is not supported in this agent version.\n"
            "Sqreen agent is disabled, you're not protected.\n"
            "Documentation can be found at %s.\n"
            "Changelog can be found at %s."
        )
        LOGGER.critical(
            msg, exception.python_version, COMPATIBILITY_URL, CHANGELOG_URL
        )
        return

    # Configure raven breadcrumbs at start
    configure_raven_breadcrumbs()

    global RUNNER_THREAD
    global RUNNER_THREAD_PID

    # Check for double instrumentation
    if RUNNER_THREAD is None:
        try:
            runner = RunnerThread(runtime_infos)
            runner.start()

            timeout = runner.instrumentation_done.wait(INSTRUMENTATION_TIMEOUT)

            if timeout is not True:
                LOGGER.critical(
                    "Sqreen couldn't start. Check network connectivity: %s", STATUS_URL)
                post_app_sqreen_exception(RemoteException(
                    "InstrumentationTimeout",
                    "Timeout during instrumentation",
                    exception_payload={
                        "runner_thread_started": runner.started,
                        "runner_thread_was_alive": runner.is_alive(),
                        "instrumentation_enabled": runner.instrumentation.enabled,
                        "runtime_infos": runtime_infos,
                    },
                    stack=runner.stacktrace,
                ).to_dict())

                # Also dump the stacktraces of all running threads
                _dump_thread_stacks()
            else:
                if runner.isAlive():
                    LOGGER.info("Sqreen instrumentation started successfully")
        except Exception:
            LOGGER.critical(
                "Sqreen thread fails to start, you're not protected",
                exc_info=True,
            )
            return

        RUNNER_THREAD = runner
        RUNNER_THREAD_PID = os.getpid()

    return RUNNER_THREAD


def get_initial_features(config_features, login_features):
    LOGGER.debug("Login features: %s", login_features)

    final_features = copy(login_features)

    if config_features:
        parsed_config_features = {}
        try:
            parsed_config_features = json.loads(config_features)
        except (ValueError, TypeError):
            LOGGER.warning(
                "Invalid config initial features %s",
                config_features,
                exc_info=True,
            )

        if parsed_config_features and isinstance(parsed_config_features, dict):
            msg = "Override login initial features with %s"
            LOGGER.warning(msg, parsed_config_features)

            final_features.update(parsed_config_features)

            LOGGER.warning("Final features %s", final_features)

    return final_features


def get_runner(
    config,
    login_result,
    session,
    queue,
    runtime_infos,
    observations_queue,
    instrumentation,
    settings,
):
    initial_features = get_initial_features(
        config["INITIAL_FEATURES"], login_result.get("features", {})
    )

    # Get the right deliverer according to initial features
    deliverer = get_deliverer(
        initial_features.get("batch_size", 0),
        initial_features.get("max_staleness", 0),
        session,
    )
    remote_command = RemoteCommand.with_production_commands()
    metrics_store = MetricsStore()
    metrics_store.register_production_aggregators()
    metrics_store.register_default_metrics()

    # Setup the runner
    runner = Runner(
        queue,
        observations_queue,
        session,
        deliverer,
        remote_command,
        runtime_infos,
        instrumentation,
        metrics_store,
        settings,
        initial_features,
    )
    return runner


class RunnerThread(Thread):
    """ Class responsible for starting the runner and monitor it
    """

    name = "SqreenRunnerThread"

    def __init__(self, runtime_infos, queue=None, observations_queue=None,
                 instrumentation=None, settings=None):
        super(RunnerThread, self).__init__()
        self.daemon = True
        self.queue = queue or CappedQueue(MAX_QUEUE_LENGTH)
        self.observations_queue = observations_queue or CappedQueue(MAX_OBS_QUEUE_LENGTH)
        self.instrumentation_done = Event()
        self.runtime_infos = runtime_infos
        self.instrumentation = instrumentation or Instrumentation(
            self.observations_queue, self.queue, before_hook_point
        )
        self.settings = settings or RunnerSettings()
        self.runner = None
        self.started = False

    @property
    def stacktrace(self):
        """Get the runner thread stack trace if it exists.
        """
        frame = sys._current_frames().get(self.ident, None)
        if frame is not None:
            return traceback.extract_stack(frame)

    def run(self):
        """ Launch the runner
        """
        self.started = True
        LOGGER.debug("Starting Sqreen %s", __version__)
        # FIXME this part should be more tested
        while True:
            session = None
            self.runner = None

            try:
                token = CONFIG["TOKEN"]
                app_name = CONFIG["APP_NAME"]
                if not token:
                    msg = (
                        "Sorry but we couldn't find your Sqreen token.\n"
                        "Your application is NOT currently protected by Sqreen.\n"
                        "\n"
                        "Have you filled your sqreen.ini?"
                    )
                    LOGGER.critical(msg)
                    self.instrumentation_done.set()
                    return

                LOGGER.warning("Using token %s", token)

                # Initiate HTTP connection to the backend and the session
                session = get_session(
                    CONFIG["URL"], token, app_name, proxy_url=CONFIG["PROXY_URL"]
                )

                runtime_infos = dict(self.runtime_infos)
                runtime_infos["various_infos"] = {
                    k: v
                    for k, v in runtime_infos.get("various_infos", {}).items()
                    if k != "dependencies"
                }
                try:
                    login_result = session.login(runtime_infos)
                except InvalidToken:
                    msg = (
                        "Sorry but your Sqreen token appears to be invalid.\n"
                        "Your application is NOT currently protected by Sqreen.\n"
                        "\n"
                        "Please check the token against the interface"
                    )
                    LOGGER.critical(msg)
                    self.instrumentation_done.set()
                    return
                except HTTPError:
                    msg = (
                        "Sorry but Sqreen's backend appears to be unavailable.\n"
                        "Your application is NOT currently protected by Sqreen.\n"
                        "\n"
                        "Please check http://sqreen.status.io for outage.\n"
                        "\n"
                        "Otherwise, back.sqreen.io may be blocked by your server's network.\n"
                        "If necessary, you tell Sqreen to connect over a proxy.\n"
                        "Please visit https://docs.sqreen.com/python/configuration/ for more details."
                    )
                    LOGGER.critical(msg)
                    self.instrumentation_done.set()
                    return

                LOGGER.info("Login success")

                self.runner = get_runner(
                    CONFIG,
                    login_result,
                    session,
                    self.queue,
                    self.runtime_infos,
                    self.observations_queue,
                    self.instrumentation,
                    self.settings,
                )

                # Process the initial commands
                process_initial_commands(login_result, self.runner)

                retry = 3
                while True:
                    try:
                        # if we did not receive instrumentation_enable, try
                        # to get it faster before the MainThread stops waiting.
                        while not self.instrumentation.enabled and retry > 0:
                            retry -= 1
                            self._random_sleep(2.0)
                            self.runner.do_heartbeat()
                        self.instrumentation_done.set()
                        self.runner.run()
                    except HTTPError:
                        LOGGER.exception(
                            "An HTTP error occured, restarting runner"
                        )
                        try:
                            session.post_sqreen_exception(
                                RemoteException.from_exc_info().to_dict()
                            )
                        except HTTPError:
                            pass
                        self._random_sleep()
                    else:
                        # If the runner exits normally, return.
                        return
            except Exception as exc:
                LOGGER.exception("An unexpected exception occured")
                if isinstance(exc, InvalidSignature):
                    remote_exception = RemoteException.from_exc_info(
                        exc.exc_info, callback_payload=exc.callback_payload  # pylint: disable=no-member
                    )
                else:
                    remote_exception = RemoteException.from_exc_info()

                if session is not None and session.is_connected():
                    try:
                        session.post_sqreen_exception(
                            remote_exception.to_dict()
                        )
                        session.logout()
                    except HTTPError:
                        LOGGER.exception("An HTTP error occured while logout")
                    except Exception:
                        LOGGER.exception("Exception while logout")
                        return
                else:
                    post_app_sqreen_exception(remote_exception.to_dict())

                try:
                    self.instrumentation.deinstrument_all()
                except Exception:
                    # We did not managed to remove instrumentation, state is unclear:
                    # terminate thread
                    LOGGER.exception("Exception while trying to clean-up")
                    return

                self._random_sleep()

    def quick_exit(self):
        """Check for early exit."""
        if os.environ.get("SQREEN_SKIP_LOGOUT", False):
            return True
        # uWSGI 2.0.15 may segfault upon exiting.
        # See https://github.com/unbit/uwsgi/issues/1651.
        uwsgi_vers = uwsgi_version_info()
        if uwsgi_vers and uwsgi_vers >= (2, 0, 15):
            return True
        elif "django" in sys.modules:
            return django_framework.DjangoRequest.DEBUG_MODE
        elif "flask" in sys.modules:
            return flask_framework.FlaskRequest.DEBUG_MODE
        elif "pyramid" in sys.modules:
            return pyramid_framework.PyramidRequest.DEBUG_MODE
        elif HAS_ASYNCIO and "aiohttp" in sys.modules:
            return aiohttp_framework.AioHTTPRequest.DEBUG_MODE
        return False

    def start(self):
        """Start the runner thread."""
        atexit.register(self.graceful_stop)
        return super(RunnerThread, self).start()

    def graceful_stop(self, force_quick_exit=False):
        """Gracefully stop the runner thread and wait for its termination."""
        self.queue.put(RunnerStop)

        if force_quick_exit or self.quick_exit():
            LOGGER.warning("Quick exit, don't logout")
            return

        try:
            self.join(GRACEFUL_TIMEOUT)
        except RuntimeError:
            logger = getLogger(__name__)
            logger.warning(GRACEFUL_FAIL_MSG, GRACEFUL_TIMEOUT)
        else:
            if self.isAlive():
                logger = getLogger(__name__)
                logger.warning(GRACEFUL_FAIL_MSG, GRACEFUL_TIMEOUT)

    def replace(self, **kwargs):
        """Create a new runner thread instance."""
        nkwargs = dict(
            queue=self.queue,
            observations_queue=self.observations_queue,
            instrumentation=self.instrumentation,
            settings=self.settings,
        )
        nkwargs.update(kwargs)
        return self.__class__(RuntimeInfos().all(), **nkwargs)

    @classmethod
    def _random_sleep(cls, max_delay=10.0):
        """Sleep for a random duration before trying again."""
        delay = random.uniform(0, max_delay)
        LOGGER.info("Sleeping %s seconds before retry", delay)
        time.sleep(delay)



# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#

import logging

from ..actions import ACTION_STORE, ActionName
from ..ip_radix import Radix
from ..rules import RuleCallback
from ..sdk import events

LOGGER = logging.getLogger(__name__)


class IPBlacklistCB(RuleCallback):
    def __init__(self, *args, **kwargs):
        super(IPBlacklistCB, self).__init__(*args, **kwargs)
        self.networks = Radix(None)
        for blacklist in self.data["values"]:
            self.networks.insert(blacklist, '/' in blacklist)
        LOGGER.debug("Blacklisted IP networks: %s", self.networks)

    def pre(self, original, *args, **kwargs):
        request = self.storage.get_current_request()
        if request is None:
            return
        client_ip = request.raw_client_ip
        if client_ip is None:
            return
        network = self.networks.match(client_ip)
        if network is not None:
            LOGGER.debug(
                "IP %s belongs to blacklisted network %s",
                client_ip,
                network,
            )
            self.record_observation("blacklisted", network, 1)
            return {
                "status": "raise",
                "data": network,
                "rule_name": self.rule_name,
            }

        # Handle security actions.
        action = ACTION_STORE.get_for_ip(client_ip)
        if not action:
            LOGGER.debug("IP %s is not blacklisted", client_ip)
            return
        events.track_action(action, {"ip_address": str(client_ip)}, storage=self.storage)
        if action.name == ActionName.BLOCK_IP:
            LOGGER.debug(
                "IP %s is blacklisted by action %s",
                client_ip,
                action.iden,
            )
            return {"status": "action_block", "action_id": action.iden}
        else:  # action.name == ActionName.REDIRECT_IP:
            LOGGER.debug(
                "IP %s is redirected to %r by action %s",
                client_ip,
                action.target_url,
                action.iden,
            )
            return {
                "status": "action_redirect",
                "action_id": action.iden,
                "target_url": action.target_url,
            }



# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Pyramid specific WSGI HTTP Request / Response stuff
"""

import os
from itertools import chain
from logging import getLogger

from ..utils import is_json_serializable
from .base import BaseRequest, BaseResponse

LOGGER = getLogger(__name__)


class PyramidRequest(BaseRequest):
    def __init__(self, request):
        super(PyramidRequest, self).__init__()
        self.request = request

        # Cache for params
        self._query_params = None
        self._query_params_values = None
        self._form_params = None
        self._json_body = None

        if self.__class__.DEBUG_MODE is None:
            self.__class__.DEBUG_MODE = self.is_debug()

    @property
    def query_params(self):
        if self._query_params is None:
            # Convert pyramid MultiDict to a normal dict with values as list
            self._query_params = dict(self.request.GET.dict_of_lists())
        return self._query_params

    @property
    def query_params_values(self):
        """ Return only query values as a list
        """
        if self._query_params_values is None:
            self._query_params_values = list(
                chain.from_iterable(self.query_params.values())
            )
        return self._query_params_values

    @property
    def form_params(self):
        # Convert pyramid MultiDict to a normal dict with values as list
        if self._form_params is None:
            form_params = {}
            post_params = self.request.POST
            for param_name in post_params:
                values = post_params.getall(param_name)
                # Ignore any non json serializable value as we don't know
                # how to process them (like cgi.FieldStorage)
                form_params[param_name] = list(
                    filter(is_json_serializable, values)
                )
            self._form_params = form_params
        return self._form_params

    @property
    def cookies_params(self):
        return dict(self.request.cookies)

    @property
    def remote_addr(self):
        """Remote IP address."""
        return self.request.remote_addr

    @property
    def hostname(self):
        return self.request.host

    @property
    def method(self):
        return self.request.method

    @property
    def referer(self):
        return self.get_raw_header("HTTP_REFERER")

    @property
    def client_user_agent(self):
        return self.request.user_agent

    @property
    def route(self):
        """Request route."""
        route = getattr(self.request, "matched_route", None)
        pattern = getattr(route, "pattern", None)
        return pattern

    @property
    def path(self):
        return self.request.path

    @property
    def scheme(self):
        return self.request.scheme

    @property
    def server_port(self):
        return self.get_raw_header("SERVER_PORT")

    @property
    def remote_port(self):
        return self.get_raw_header("REMOTE_PORT")

    @property
    def view_params(self):
        return self.request.matchdict

    @property
    def json_params(self):
        if self._json_body is None:
            try:
                self._json_body = self.request.json_body
            except Exception:
                self._json_body = {}
        return self._json_body

    @property
    def raw_headers(self):
        return self.request.environ

    @staticmethod
    def is_debug():
        try:
            return bool(os.environ.get("PYTHON_RELOADER_SHOULD_RUN", False))
        except Exception:
            LOGGER.warning("Exception when checking debug mode", exc_info=1)
            return False


class PyramidResponse(BaseResponse):

    def __init__(self, response):
        self.response = response

    @property
    def status_code(self):
        return self.response.status_code

    @property
    def content_type(self):
        return self.response.headers.get("Content-Type")

    @property
    def content_length(self):
        try:
            return int(self.response.content_length)
        except (ValueError, TypeError):
            return None



# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
"""Patch an async event loop class with a context-aware task factory."""

import logging

from ...async_context import create_task_factory
from .import_hook import ImportHookStrategy

LOGGER = logging.getLogger(__name__)


def _patch_loop_cls(loop_cls):
    """Patch an async event loop class with a context-aware task factory.

    Return the modified class constructor.
    """
    orig_init = loop_cls.__init__
    orig_set_task_factory = loop_cls.set_task_factory

    def __init__(self, *args, **kwargs):
        orig_init(self, *args, **kwargs)
        self.set_task_factory(None)

    def set_task_factory(self, factory):
        factory = create_task_factory(factory)
        orig_set_task_factory(self, factory)

    setattr(loop_cls, "__init__", __init__)
    setattr(loop_cls, "set_task_factory", set_task_factory)
    return loop_cls


class AsyncEventLoopStrategy(ImportHookStrategy):
    """Patch an async event loop class with a context-aware task factory."""

    def import_hook_callback(self, original):
        return _patch_loop_cls(original)



# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
"""Insert custom headers for aiohttp."""

import logging

from .headers_insert import BaseHeadersInsertCB

LOGGER = logging.getLogger(__name__)


class HeadersInsertCBAioHTTP(BaseHeadersInsertCB):
    def post(self, original, response, *args, **kwargs):
        """Insert custom headers."""
        try:
            for header_name, header_value in self.headers.items():
                response.headers[header_name] = header_value
        except Exception:
            LOGGER.warning("An error occured", exc_info=True)

        return {}



# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Count http codes for Django framework
"""
import logging

from ..rules import RuleCallback

LOGGER = logging.getLogger(__name__)


class CountHTTPCodesCBDjango(RuleCallback):
    def post(self, original, response, *args, **kwargs):
        """ Recover the status code and update the http_code metric
        """
        status_code = response.status_code
        self.record_observation("http_code", str(status_code), 1)

        return {}



# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Record the current request in flask application
"""
from logging import getLogger

from ..frameworks.flask_framework import FlaskRequest, FlaskResponse
from .record_request_context import RecordRequestContext

LOGGER = getLogger(__name__)


class RecordRequestContextFlask(RecordRequestContext):
    def pre(self, original, *args, **kwargs):
        from flask import request

        self._store_request(FlaskRequest(request))

    def post(self, *args, **kwargs):
        self.storage.store_response(FlaskResponse(args[1]))
        self._clear_request()



# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Count things according to a list of binding accessors expression
"""
import json
import logging
from collections import Mapping

from ..binding_accessor import BindingAccessor
from ..exceptions import InvalidArgument
from ..rules import RuleCallback

LOGGER = logging.getLogger(__name__)


class BindingAccessorCounter(RuleCallback):
    def __init__(self, *args, **kwargs):
        super(BindingAccessorCounter, self).__init__(*args, **kwargs)

        if not isinstance(self.data, Mapping):
            msg = "Invalid data type received: {}"
            raise InvalidArgument(msg.format(type(self.data)))

        self.patterns = [BindingAccessor(exp) for exp in self.data["values"]]
        self.metric_name = self.metrics[0]["name"]

    def post(self, original, response, *args, **kwargs):
        """ Resolve binding expressions with the HTTP Response and send
        an observation for the rule metric.
        """
        request = self.storage.get_current_request()

        binding_eval_args = {
            "binding": locals(),
            "global_binding": globals(),
            "framework": request,
            "instance": original,
            "arguments": self.storage.get_current_args(args),
            "kwarguments": kwargs,
            "cbdata": self.data,
            "return_value": response,
        }

        key = [
            binding.resolve(**binding_eval_args) for binding in self.patterns
        ]
        formatted_key = json.dumps(key, separators=(",", ":"))
        self.record_observation(self.metric_name, formatted_key, 1)

        return {}

    def failing(self, original, *args, **kwargs):
        """ Resolve binding expressions with the exception, replace the status
        code with 500 and send an observation for the rule metric.
        """
        request = self.storage.get_current_request()

        binding_eval_args = {
            "binding": locals(),
            "global_binding": globals(),
            "framework": request,
            "instance": original,
            "arguments": self.storage.get_current_args(args),
            "kwarguments": kwargs,
            "cbdata": self.data,
            "return_value": None,
        }

        key = []
        for binding in self.patterns:

            # Replace status code replacement by 500 as we don't have a response
            if binding.expression.endswith(
                ".status_code"
            ) or binding.expression.endswith(".status"):
                key.append(500)
                continue

            key.append(binding.resolve(**binding_eval_args))
        formatted_key = json.dumps(key, separators=(",", ":"))

        self.record_observation(self.metric_name, formatted_key, 1)

        return {}



# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
"""Custom error page for aiohttp."""

from logging import getLogger

from ..exceptions import RequestBlocked
from .headers_insert import convert_to_str
from .sqreen_error_page import BaseSqreenErrorPage

LOGGER = getLogger(__name__)


class SqreenErrorPageAioHTTP(BaseSqreenErrorPage):
    """Custom error page for aiohttp."""

    @staticmethod
    def _get_exception(handler, request, status=500, exc=None, message=None):
        if isinstance(exc, RequestBlocked):
            return exc

    @staticmethod
    def _get_response(exc, content, status_code, headers={}):
        from aiohttp.web_response import Response

        resp = Response(status=status_code, text=content)
        for header_name, header_value in convert_to_str(headers.items()):
            resp.headers[header_name] = header_value
        return resp



# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
from setuptools import find_packages, setup

metadata = {}
with open("sqreen/__about__.py") as fp:
    exec(fp.read(), metadata)


long_description = """Sqreen is a SaaS based Application protection and monitoring platform that integrates directly into your Python applications.
Learn more at `<https://www.sqreen.com/>`_."""


setup(
    name="sqreen",
    version=metadata["__version__"],
    description="Sqreen agent to protect Python applications.",
    long_description=long_description,
    author=metadata["__author__"],
    author_email=metadata["__email__"],
    url="https://www.sqreen.com/",
    packages=find_packages(
        exclude=[
            "benchmarks.*", "benchmarks",
            "tests.*", "tests",
        ]
    ),
    package_dir={"sqreen": "sqreen"},
    package_data={"sqreen": ["*.crt", "rules_callbacks/*.html"]},
    include_package_data=True,
    install_requires=["py-mini-racer>=0.1.18", "sq-native>=0.2,<1"],
    license=metadata["__license__"],
    zip_safe=False,
    keywords="sqreen",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: Other/Proprietary License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Security",
    ],
    test_suite="tests",
    tests_require=[],
    entry_points={
        "console_scripts": ["sqreen-start = sqreen.bin.protect:protect"]
    },
)



# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#

import logging
import uuid

from ..binding_accessor import BindingAccessor
from ..exceptions import InvalidArgument
from ..rules import RuleCallback

try:
    from sq_native import waf, lib
except ImportError:
    lib = None


LOGGER = logging.getLogger(__name__)


class WAFCB(RuleCallback):
    """ WAF Callback
    """

    def __init__(self, *args, **kwargs):
        super(WAFCB, self).__init__(*args, **kwargs)
        if not self.lib_is_available():
            msg = "WAF will be disabled because the native module is absent"
            raise InvalidArgument(msg)

        prefix = uuid.uuid4()
        waf_rule_name = "{}_{}".format(prefix, self.rule_name)

        values = self.data.get("values", {})
        ret = waf.initialize(waf_rule_name, values.get("waf_rules"))
        if ret is not True:
            msg = "failed to initialize the WAF rules"
            raise InvalidArgument(msg)
        self._waf_rule_name = waf_rule_name

        bas = {}
        for expr in values.get("binding_accessors", []):
            bas[expr] = BindingAccessor(expr)
        self.binding_accessors = bas
        self._budget = values.get("budget", 5000)

    def __del__(self):
        waf_rule_name = getattr(self, "_waf_rule_name", None)
        if waf_rule_name is not None:
            waf.clear(waf_rule_name)

    def pre(self, original, *args, **kwargs):
        waf_rule_name = getattr(self, "_waf_rule_name", None)
        request = self.storage.get_current_request()
        if waf_rule_name is None or request is None:
            return

        binding_eval_args = {
            "binding": locals(),
            "global_binding": globals(),
            "framework": request,
            "instance": original,
            "arguments": self.storage.get_current_args(args),
            "kwarguments": kwargs,
            "cbdata": self.data,
            "return_value": None,
        }

        params = {}
        for expr, ba in self.binding_accessors.items():
            params[expr] = ba.resolve(**binding_eval_args)

        LOGGER.debug("WAF run with parameters: %r", params)
        ret = waf.run(waf_rule_name, params, self._budget)
        LOGGER.debug("WAF returned: %r", ret)
        if ret.action in (waf.PW_BLOCK, waf.PW_MONITOR):
            self.record_attack({"waf_data": ret.data})
            if ret.action == waf.PW_BLOCK and self.block is True:
                return {"status": "raise", "rule_name": self.rule_name}
        elif ret.action < 0:
            LOGGER.debug("WAF returned unexpected action: %r", ret)

    @staticmethod
    def lib_is_available():
        """ Return True if the native module is present and loaded, False otherwise.
        """
        return lib is not None



# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Custom error page for Django
"""
from ..exceptions import RequestBlocked
from .headers_insert import convert_to_str
from .sqreen_error_page import BaseSqreenErrorPage


class SqreenErrorPageDjango(BaseSqreenErrorPage):
    @staticmethod
    def _get_exception(handler, request, resolved, exc_infos, *args, **kwargs):
        if not exc_infos:
            return

        exc_class = exc_infos[0]

        # Ignore exception which are not attack blocked
        if not issubclass(exc_class, RequestBlocked):
            return

        return exc_infos[1]

    @staticmethod
    def _get_response(exception, content, status_code, headers=None):
        from django.http import HttpResponse

        response = HttpResponse(content)
        response.status_code = status_code

        if headers:
            for header_name, header_value in convert_to_str(headers.items()):
                response[header_name] = header_value

        return response



# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Base Request class
"""

import logging
import traceback
import uuid
from collections import defaultdict

from ..config import CONFIG
from ..remote_exception import traceback_formatter
from ..utils import to_unicode_safe
from .ip_utils import get_real_user_ip

LOGGER = logging.getLogger(__name__)


# Header name that we need to sends back on attack in order to compute the
# real user ip
CLIENT_IP_HEADERS = (
    "HTTP_X_FORWARDED_FOR",
    "X_FORWARDED_FOR",
    "HTTP_X_REAL_IP",
    "HTTP_CLIENT_IP",
    "HTTP_X_FORWARDED",
    "HTTP_X_CLUSTER_CLIENT_IP",
    "HTTP_FORWARDED_FOR",
    "HTTP_FORWARDED",
    "HTTP_VIA",
    "REMOTE_ADDR",
)


class BaseRequest(object):

    DEBUG_MODE = None

    def __init__(self, storage=None):
        self._flat_request_params = None
        self._flat_request_params_keys = None
        self._request_headers = None
        self.request_id = uuid.uuid4().hex
        self.storage = storage

        if CONFIG["IP_HEADER"]:
            self._client_ip_headers = [
                "HTTP_%s" % CONFIG["IP_HEADER"].upper().replace("-", "_")
            ] + list(CLIENT_IP_HEADERS)
        else:
            self._client_ip_headers = CLIENT_IP_HEADERS

    @property
    def remote_addr(self):
        raise NotImplementedError

    @property
    def hostname(self):
        raise NotImplementedError

    @property
    def method(self):
        raise NotImplementedError

    @property
    def referer(self):
        raise NotImplementedError

    @property
    def client_user_agent(self):
        raise NotImplementedError

    @property
    def path(self):
        raise NotImplementedError

    @property
    def scheme(self):
        raise NotImplementedError

    @property
    def server_port(self):
        raise NotImplementedError

    @property
    def remote_port(self):
        raise NotImplementedError

    @property
    def raw_headers(self):
        raise NotImplementedError

    @property
    def form_params(self):
        raise NotImplementedError

    @property
    def query_params(self):
        raise NotImplementedError

    @property
    def view_params(self):
        raise NotImplementedError

    @property
    def json_params(self):
        raise NotImplementedError

    @property
    def cookies_params(self):
        raise NotImplementedError

    @property
    def body(self):
        """Return the raw body of the request. By default it will not be present"""
        return None

    @property
    def request_payload(self):
        """ Returns current request payload with the backend expected field
        name. All fields should be serializable to JSON.
        """
        return {
            "rid": self.request_id,
            "remote_ip": self.remote_addr,
            "client_ip": self.client_ip,
            "host": self.hostname,
            "verb": self.method,
            "referer": self.referer,
            "user_agent": self.client_user_agent,
            "path": self.path,
            "scheme": self.scheme,
            "port": self.server_port,
            "remote_port": self.remote_port,
        }

    @property
    def request_params(self):
        """ Returns all the inputs that are controllable by the user.
        All fields should be serializable to JSON.
        """
        return {
            "form": self.form_params,
            "query": self.query_params,
            "other": self.view_params,
            "json": self.json_params,
            "cookies": self.cookies_params,
        }

    @property
    def request_params_list(self):
        """ Returns all the inputs that are controllable by the user without
        keys. All fields should be serializable to JSON.
        """
        return [
            self.form_params,
            self.query_params,
            self.view_params,
            self.json_params,
            self.cookies_params,
        ]

    def _flatten_request_params(self):
        """ Flatten the request params to compute the values and keys for
        quick access
        """
        iteration = 0
        max_iterations = 100

        values = set()
        keys = set()
        remaining_iterables = list(self.request_params_list)

        while len(remaining_iterables) != 0:

            iteration += 1
            # If we have a very big or nested iterable, returns False
            if iteration >= max_iterations:
                break

            iterable_value = remaining_iterables.pop(0)

            # If we get an iterable, add it to the list of remaining
            if isinstance(iterable_value, dict):
                dict_values = iterable_value.values()

                # Be sure to not extend with an empty dict, faster check
                if len(dict_values) > 0:
                    remaining_iterables.extend(list(dict_values))

                dict_keys = iterable_value.keys()

                if len(dict_keys) > 0:
                    keys.update(dict_keys)

            elif isinstance(iterable_value, list):
                # Be sure to not extend with an empty list, faster check
                if len(iterable_value) > 0:
                    remaining_iterables.extend(iterable_value)
            else:
                values.add(iterable_value)

        self._flat_request_params = values
        self._flat_request_params_keys = keys

    @property
    def flat_request_params(self):
        """ Return only the request params values for all request params
        """
        if self._flat_request_params is None:
            self._flatten_request_params()

        return self._flat_request_params

    @property
    def flat_request_params_keys(self):
        """ Return only the request params names for all request params
        """
        if self._flat_request_params_keys is None:
            self._flatten_request_params()
        return self._flat_request_params_keys

    def params_contains(self, param):
        """ Return True if the parameter given in input is present in the
        request inputs.
        """
        return param in self.flat_request_params

    @property
    def flat_cookies_values(self):
        """ Return the cookies values
        """
        return list(self.cookies_params.values())

    @property
    def flat_cookies_keys(self):
        """ Return the cookies values
        """
        return list(self.cookies_params.keys())

    @property
    def request_headers(self):
        if self._request_headers is None:
            self._request_headers = defaultdict(str)

            for key, value in self.raw_headers.items():

                if not key.startswith("HTTP_"):
                    if key not in ("CONTENT_TYPE", "CONTENT_LENGTH"):
                        continue

                self._request_headers[to_unicode_safe(key)] = to_unicode_safe(value)
        return self._request_headers

    @property
    def content_type(self):
        return self.get_raw_header("CONTENT_TYPE")

    @property
    def content_length(self):
        try:
            return int(self.get_raw_header("CONTENT_LENGTH"))
        except ValueError:
            return None

    @property
    def flat_headers_values(self):
        return list(self.request_headers.values())

    @property
    def flat_headers_keys(self):
        return list(self.request_headers.keys())

    @property
    def full_payload(self):
        """ Return request information in the format accepted by the backend
        """
        return {"request": self.request_payload, "params": self.request_params}

    @property
    def caller(self):
        return traceback.format_stack()

    @property
    def raw_caller(self):
        return traceback_formatter(traceback.extract_stack())

    @property
    def client_ip(self):
        """String formatted client IP address."""
        ip = self.raw_client_ip
        if ip is not None:
            return to_unicode_safe(ip)

    @property
    def raw_client_ip(self):
        """Client IP address."""
        return get_real_user_ip(self.remote_addr, *self.iter_client_ips())

    def get_raw_header(self, name, default=None):
        """ Get a specific raw header."""
        return to_unicode_safe(self.raw_headers.get(name, default))

    def iter_client_ips(self):
        """Yield the client IPs set in raw headers."""
        for header_name in self._client_ip_headers:
            value = self.get_raw_header(header_name)
            if value:
                yield value

    def get_client_ips_headers(self):
        """ Return raw headers that can be used to find the real client ip
        """
        raw_headers = []
        for header_name in self._client_ip_headers:
            value = self.get_raw_header(header_name)
            if value:
                raw_headers.append([header_name, value])

        return raw_headers


class BaseResponse(object):

    @property
    def status_code(self):
        raise NotImplementedError

    @property
    def content_type(self):
        raise NotImplementedError

    def content_length(self):
        raise NotImplementedError

    @property
    def response_payload(self):
        """Returns the response information understood by the backend.
        """
        return {
            "status": self.status_code,
            "content_type": self.content_type,
            "content_length": self.content_length,
        }



# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
"""Sanitizer used to remove sensitive data from our payload"""

import logging
import re

from . import config
from .utils import is_string, is_unicode

LOGGER = logging.getLogger(__name__)

MASK = '<Redacted by Sqreen>'

SENSITIVE_KEYS = [k.strip() for k in config.CONFIG["STRIP_SENSITIVE_KEYS"].split(',')]
LOGGER.debug("Using sensitive keys %s", ", ".join(SENSITIVE_KEYS))
try:
    SENSITIVE_REGEX = re.compile(config.CONFIG["STRIP_SENSITIVE_REGEX"])
except re.error:
    SENSITIVE_REGEX = re.compile(config.CONFIG_DEFAULT_VALUE["STRIP_SENSITIVE_REGEX"])
    LOGGER.warning("Invalid regexp configuration: %s", config.CONFIG["STRIP_SENSITIVE_REGEX"])
finally:
    LOGGER.debug("Using sensitive regex %s", SENSITIVE_REGEX.pattern)


def strip_sensitive_data(data, keys=SENSITIVE_KEYS, regex=SENSITIVE_REGEX):
    """Sanitize sensitive data from an object"""
    if is_string(data):
        if not is_unicode(data):
            data = data.decode("utf-8", errors="replace")

        if regex.match(data):
            return MASK
        return data

    if isinstance(data, dict):
        sanitize_data = {}
        for k, v in data.items():
            if k in keys:
                sanitize_data[k] = MASK
            else:
                sanitize_data[k] = strip_sensitive_data(v, keys, regex)

        return sanitize_data

    if isinstance(data, (list, tuple, set)):
        return [strip_sensitive_data(v, keys, regex) for v in data]

    return data



# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Look for known crawlers user-agents
"""
from logging import getLogger

from ..frameworks.django_framework import DjangoRequest, DjangoResponse
from .record_request_context import RecordRequestContext

LOGGER = getLogger(__name__)


class RecordRequestContextDjango(RecordRequestContext):
    def pre(self, original, request, view_func, view_args, view_kwargs):
        self._store_request(
            DjangoRequest(request, view_func, view_args, view_kwargs)
        )

    def post(self, *args, **kwargs):
        self.storage.store_response(DjangoResponse(args[1]))
        self._clear_request()

    @staticmethod
    def failing(*args, **kwargs):
        """ Post is always called in a Django Middleware, don't clean the
        request right now as it may be needed in a post callback
        """
        pass



# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Contains the hook_point, the code actually executed in place of
application code
"""
import logging
import sys
from collections import Iterable, Mapping
from traceback import extract_stack

from ..constants import ACTIONS, LIFECYCLE_METHODS, VALID_ACTIONS_PER_LIFECYCLE
from ..exceptions import (
    ActionBlock,
    ActionRedirect,
    AttackBlocked,
    MissingDataException,
    SqreenException,
)
from ..remote_exception import RemoteException
from ..runtime_storage import runtime
from ..utils import update_wrapper
from .helpers import guard_call

LOGGER = logging.getLogger(__name__)


def _compute_result_action(current_result, next_result, valid_actions):
    """ Compute the next result_action based on current one and the callback
    result
    """
    current_status = current_result.get("status")
    next_status = next_result.get("status")
    # Check the validity of action per method
    if next_status not in valid_actions:
        # Ignore it
        return current_result

    if current_status not in valid_actions:
        return next_result

    current_result_order = valid_actions.index(current_result.get("status"))
    next_result_order = valid_actions.index(next_result.get("status"))

    # Return the old one only if it has more priority
    if current_result_order < next_result_order:
        return current_result
    else:
        return next_result


def valid_args(new_args):
    """ Check that new_args match the format ([], {}), else returns None
    """
    try:
        args, kwargs = new_args
    except (TypeError, ValueError):
        msg = "Invalid number of args or type, %s"
        LOGGER.debug(msg, new_args)
        return False

    # Check that new_args[0] is a valid type for passing to *
    if not isinstance(args, Iterable) or isinstance(args, Mapping):
        msg = "Invalid type for args[0]: %s"
        LOGGER.debug(msg, args)
        return False

    # Check that new_args[1] is a valid type for passing to **
    if not isinstance(kwargs, Mapping):
        msg = "Invalid type for args[1]: %s"
        LOGGER.debug(msg, kwargs)
        return False

    return True


def execute_callback(
    queue, callback, method, original, result_action, args, kwargs,
    storage=runtime, budget_in_ms=None
):
    """ Execute a callback protected inside a broad try except.

    The queue argument is used to push a RemoteException when an exception is raised by the callback.
    The result_action argument is a dict where each callback pushes its results.
    """

    callback_method = getattr(callback, method)
    callback_rule_name = callback.rule_name
    LOGGER.debug(
        "Running %s callback %s of module '%s'",
        method,
        callback_method,
        callback_method.__module__,
    )

    if args is None:
        args = tuple()

    if kwargs is None:
        kwargs = {}

    try:
        if callback.supports_budget:
            result = callback_method(original, __sqreen_budget=budget_in_ms, *args, **kwargs)
        else:
            result = callback_method(original, *args, **kwargs)

        # First process of result

        if not result:
            return result_action, args, kwargs

        if not callback.block:
            LOGGER.debug("%s cannot block, ignoring return value", callback)
            return result_action, args, kwargs

        if not isinstance(result, dict):
            msg = "%s returns %s during %s method"
            LOGGER.warning(msg, callback, type(result), method)
            return result_action, args, kwargs

        # Set rule_name if not set
        if result.get("rule_name") is None:
            result["rule_name"] = callback_rule_name

        valid_actions = VALID_ACTIONS_PER_LIFECYCLE[method]
        result_action = _compute_result_action(
            result_action, result, valid_actions
        )

        # We need to process MODIFY_ARGS for next callbacks
        if result_action.get("status") == ACTIONS["MODIFY_ARGS"]:
            if valid_args(result_action["args"]):
                args, kwargs = result_action["args"]

        return result_action, args, kwargs

    except MissingDataException:
        raise

    except Exception as exception:
        LOGGER.exception(
            "An exception occured while trying to execute %s", callback
        )

        exc_info = sys.exc_info()
        current_request = storage.get_current_request()
        stack = extract_stack()

        if current_request:
            callback.record_exception(exception, exc_info, stack)
        else:
            callback_exception_payload = callback.exception_infos()

            # Try to recover some infos from the exception if it's a SqreenException
            exception_infos = {}
            if isinstance(exception, SqreenException):
                exception_infos.update(exception.exception_infos())  # pylint: disable=no-member

            remote_exception = RemoteException.from_exc_info(
                exc_info,
                callback_payload=callback_exception_payload,
                exception_payload=exception_infos,
                stack=stack,
            )
            queue.put(remote_exception)

        return result_action, args, kwargs


def execute_callbacks(
    queue, callbacks, method, original, args=None, kwargs=None, storage=runtime, budget=None,
):
    """ Execute a list of callbacks method (pre/post/fail), catch any exception
    that could happens.

    Aggregate the callbacks result (format {"status": COMMAND}), compute the
    final ACTION to execute and return it.
    For every ACTION, each callback is executed even if the first callback
    detected an attack.
    """
    result_action = {}

    # Execute the callbacks
    for callback in callbacks:
        if callback.whitelisted:
            continue

        if budget == 0:
            break

        with storage.trace() as callback_trace:
            result = execute_callback(
                queue, callback, method, original, result_action, args, kwargs,
                storage=storage, budget_in_ms=budget
            )

        if budget is not None:
            budget -= callback_trace.duration_ms
            if budget < 0:
                budget = 0

        result_action, args, kwargs = result

    # It is either empty dict if no override or the last one
    return result_action


def execute_pre_callbacks(key, strategy, original, args=None, kwargs=None, storage=runtime, budget=None):
    """ Execute pre_callbacks. Pre callbacks will receive these arguments:
    (original, *args, **kwargs)
    """
    pre_callbacks = strategy.get_pre_callbacks(key)

    if len(pre_callbacks) > 0:
        return guard_call(
            key,
            execute_callbacks,
            strategy.queue,
            pre_callbacks,
            LIFECYCLE_METHODS["PRE"],
            original,
            args,
            kwargs,
            storage=storage,
            budget=budget,
        )
    else:
        return {}


def execute_failing_callbacks(
    key, strategy, original, exc_infos, args=None, kwargs=None, storage=runtime, budget=None
):
    """ Execute failing_callbacks. Failing callbacks will receive these arguments:
    (original, exc_infos, *args, **kwargs)
    """
    failing_callbacks = strategy.get_failing_callbacks(key)

    if len(failing_callbacks) > 0:
        if args is None:
            args = tuple()

        new_args = [exc_infos] + list(args)

        return guard_call(
            key,
            execute_callbacks,
            strategy.queue,
            reversed(failing_callbacks),
            LIFECYCLE_METHODS["FAILING"],
            original,
            new_args,
            kwargs,
            storage=storage,
            budget=budget,
        )
    else:
        return {}


def execute_post_callbacks(
    key, strategy, original, result, args=None, kwargs=None, storage=runtime, budget=None
):
    """ Execute post_callbacks. Post callbacks will receive these arguments:
    (original, result, *args, **kwargs)
    """
    post_callbacks = strategy.get_post_callbacks(key)

    if len(post_callbacks) > 0:
        if args is None:
            args = tuple()

        new_args = [result] + list(args)

        return guard_call(
            key,
            execute_callbacks,
            strategy.queue,
            reversed(post_callbacks),
            LIFECYCLE_METHODS["POST"],
            original,
            new_args,
            kwargs,
            storage=storage,
            budget=budget
        )
    else:
        return {}


def hook_point(strategy, hook_name, hook_method, original, storage=runtime):
    """ Execute the original method and pre/post/failing callbacks
    If an exception happens, create a RemoteException, call
    callback.exception_infos for more debugging infos and send it via
    the provided queue.
    """

    def wrapper(*args, **kwargs):
        LOGGER.info(
            "Checking before hook point of %s for %s/%s",
            strategy,
            hook_name,
            hook_method,
        )
        strategy.before_hook_point()
        key = (hook_name, hook_method)

        # Call pre callbacks
        action = execute_pre_callbacks(
            key,
            strategy,
            original,
            args,
            kwargs,
            storage=storage
        )

        if action.get("status") == ACTIONS["RAISE"]:
            LOGGER.debug(
                "Callback %s detected an attack", action.get("rule_name")
            )
            raise AttackBlocked(action.get("rule_name"))
        elif action.get("status") == ACTIONS["ACTION_BLOCK"]:
            LOGGER.debug(
                "Action %s blocked the request", action.get("action_id")
            )
            raise ActionBlock(action.get("action_id"))
        elif action.get("status") == ACTIONS["ACTION_REDIRECT"]:
            LOGGER.debug(
                "Action %s redirected the request to %r",
                action.get("action_id"),
                action["target_url"],
            )
            raise ActionRedirect(action.get("action_id"), action["target_url"])
        elif action.get("status") == ACTIONS["OVERRIDE"]:
            return action.get("new_return_value")
        elif action.get("status") == ACTIONS["MODIFY_ARGS"]:
            if valid_args(action["args"]):
                args, kwargs = action["args"]

        # Call original method
        retry = True
        while retry is True:
            try:
                retry = False
                # Try to call original function
                result = original(*args, **kwargs)
            except Exception:
                # In case of error, call fail callbacks with exception infos
                exc_infos = sys.exc_info()

                # Either raise an exception, set a return value or retry
                action = execute_failing_callbacks(
                    key, strategy, original, exc_infos, args, kwargs,
                    storage=storage
                )

                if action.get("status") == ACTIONS["RAISE"]:
                    LOGGER.debug(
                        "Callback %s detected an attack",
                        action.get("rule_name"),
                    )
                    raise AttackBlocked(action.get("rule_name"))
                elif action.get("status") == ACTIONS["RETRY"]:
                    retry = True
                elif action.get("status") == ACTIONS["OVERRIDE"]:
                    return action.get("new_return_value")

                # Be sure to raise if no retry or override
                if retry is False:
                    raise

        # Then call post callback in reverse order to simulate decorator
        # behavior
        action = execute_post_callbacks(
            key, strategy, original, result, args, kwargs, storage=storage,
        )

        if action.get("status") == ACTIONS["RAISE"]:
            LOGGER.debug(
                "Callback %s detected an attack", action.get("rule_name")
            )
            raise AttackBlocked(action.get("rule_name"))
        elif action.get("status") == ACTIONS["OVERRIDE"]:
            return action.get("new_return_value")

        # And return the original value
        return result

    # Update wrapper name, module and doc
    update_wrapper(wrapper, original)

    return wrapper



# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Package for framework middlewares integrations
"""
from ...utils import HAS_ASYNCIO
from .django_middleware import DjangoMiddleware
from .flask_middleware import FlaskMiddleware
from .pyramid_middleware import PyramidMiddleware

if HAS_ASYNCIO:
    from .aiohttp_middleware import AioHTTPMiddleware

__all__ = [
    "AioHTTPMiddleware",
    "DjangoMiddleware",
    "FlaskMiddleware",
    "PyramidMiddleware",
]



# -*- coding: utf-8 -*-
# Copyright (c) 2016, 2017, 2018, 2019 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Custom error page for Pyramid
"""
from ..exceptions import RequestBlocked
from .headers_insert import convert_to_str
from .sqreen_error_page import BaseSqreenErrorPage


class SqreenErrorPagePyramid(BaseSqreenErrorPage):
    @staticmethod
    def _get_exception(registry, request, context, *args, **kwargs):
        if not context:
            return

        # Ignore exception which are not attack blocked
        if not isinstance(context, RequestBlocked):
            return

        return context

    @staticmethod
    def _get_response(exception, content, status_code, headers={}):
        from pyramid.response import Response

        response = Response(content, status_code=status_code)

        for header_name, header_value in convert_to_str(headers.items()):
            response.headers[header_name] = header_value

        return response
